from __future__ import print_function, division
from coffea import processor

import multiprocessing
import sys
import pytest


def test_parsl_start_stop():
    parsl = pytest.importorskip("parsl", minversion="0.7.2")

    from coffea.processor.parsl.detail import (_parsl_initialize,
                                               _parsl_stop,
                                               _default_cfg)

    _parsl_initialize(config=_default_cfg)

    _parsl_stop()


def do_parsl_job(filelist, flatten=False, compression=0, config=None):
    treename='Events'

    from coffea.processor.test_items import NanoTestProcessor
    proc = NanoTestProcessor()

    exe_args = {
        'flatten': flatten,
        'compression': compression,
        'config': config,
    }
    hists = processor.run_parsl_job(filelist,
                                    treename,
                                    processor_instance=proc,
                                    executor=processor.executor.parsl_executor,
                                    executor_args=exe_args)

    assert( hists['cutflow']['ZJets_pt'] == 18 )
    assert( hists['cutflow']['ZJets_mass'] == 6 )
    assert( hists['cutflow']['Data_pt'] == 84 )
    assert( hists['cutflow']['Data_mass'] == 66 )


def test_parsl_htex_executor():
    parsl = pytest.importorskip("parsl", minversion="0.7.2")
    import os
    import os.path as osp

    from parsl.providers import LocalProvider
    from parsl.channels import LocalChannel
    from parsl.executors import HighThroughputExecutor
    from parsl.addresses import address_by_hostname
    from parsl.config import Config
    parsl_config = Config(
        executors=[
            HighThroughputExecutor(
                label="coffea_parsl_default",
                address=address_by_hostname(),
                cores_per_worker=max(multiprocessing.cpu_count()//2, 1),
                max_workers=1,
                provider=LocalProvider(
                    channel=LocalChannel(),
                    init_blocks=1,
                    max_blocks=1,
                    nodes_per_block=1
                ),
            )
        ],
        strategy=None,
    )
    parsl.load(parsl_config)

    filelist = {
        'ZJets': [osp.join(os.getcwd(),'tests/samples/nano_dy.root')],
        'Data' : [osp.join(os.getcwd(),'tests/samples/nano_dimuon.root')]
    }

    do_parsl_job(filelist)
    do_parsl_job(filelist, compression=1)

    filelist = {
        'ZJets': {'treename': 'Events', 'files': [osp.join(os.getcwd(),'tests/samples/nano_dy.root')]},
        'Data' : {'treename': 'Events', 'files': [osp.join(os.getcwd(),'tests/samples/nano_dimuon.root')]}
    }

    do_parsl_job(filelist)
    do_parsl_job(filelist, flatten=True)


def test_parsl_deprecated():
    parsl = pytest.importorskip("parsl", minversion="0.7.2")
    
    from coffea.processor.parsl.parsl_executor import coffea_pyapp_func, parsl_executor

    with pytest.raises(RuntimeError):
        coffea_pyapp_func(None, None, None, None, None, None)

    with pytest.raises(RuntimeError):
        parsl_executor(None, None, None, None)


@pytest.mark.skipif(sys.platform.startswith('win'), reason='signals are different on windows')
def test_timeout():
    from coffea.processor.parsl.timeout import timeout
    import signal

    @timeout
    def too_long(timeout=None):
        import time
        time.sleep(20)

    @timeout
    def make_except(timeout=None):
        import time
        time.sleep(1)
        raise Exception('oops!')

    try:
        too_long(timeout=5)
    except Exception as e:
        assert(e.args[0] == "Timeout hit")

    try:
        make_except(timeout=20)
    except Exception as e:
        assert(e.args[0] == 'oops!')

    # reset alarms for other tests, this is suspicious
    signal.alarm(0)


def test_parsl_condor_cfg():
    parsl = pytest.importorskip("parsl", minversion="0.7.2")

    from coffea.processor.parsl.condor_config import condor_config

    test = condor_config()


def test_parsl_slurm_cfg():
    import os
    parsl = pytest.importorskip("parsl", minversion="0.7.2")

    x509_proxy = 'x509up_u%s' % (os.getuid())
    fname = '/tmp/%s' % x509_proxy
    with open(fname, 'w+'):
        os.utime(fname, None)

    from coffea.processor.parsl.slurm_config import slurm_config

    test = slurm_config()



import os
import blosc
from threading import Lock
from collections.abc import MutableMapping
from distributed import WorkerPlugin, get_worker
from zict import Buffer, Func, LRU, File


class ColumnCache(WorkerPlugin, MutableMapping):
    name = 'columncache'

    def __init__(self, maxmem=5e8, maxcompressed=2e9, maxdisk=1e10):
        self._maxmem = maxmem
        self._maxcompressed = maxcompressed
        self._maxdisk = maxdisk

    def setup(self, worker):
        self.cache = Buffer(
            fast={},
            slow=Func(
                dump=blosc.pack_array,
                load=blosc.unpack_array,
                d=Buffer(
                    fast={},
                    slow=LRU(
                        n=self._maxdisk,
                        d=File(os.path.join(worker.local_directory, 'cache')),
                        weight=lambda k, v: len(v),
                    ),
                    n=self._maxcompressed,
                    weight=lambda k, v: len(v),
                ),
            ),
            n=self._maxmem,
            weight=lambda k, v: v.nbytes,
        )
        self.lock = Lock()
        self.hits = 0
        self.misses = 0

    def teardown(self, worker):
        pass

    def __getitem__(self, key):
        with self.lock:
            try:
                out = self.cache[key]
                self.hits += 1
                return out
            except KeyError:
                self.misses += 1
                raise

    def __setitem__(self, key, value):
        with self.lock:
            self.cache[key] = value

    def __delitem__(self, key):
        with self.lock:
            del self.cache[key]

    def __iter__(self):
        with self.lock:
            return iter(self.cache)

    def __len__(self):
        with self.lock:
            return len(self.cache)


def register_columncache(client):
    plugins = set()
    for p in client.run(lambda: set(get_worker().plugins)).values():
        plugins |= p
    if ColumnCache.name not in plugins:
        client.register_worker_plugin(ColumnCache())



from __future__ import print_function, division
import concurrent.futures
from functools import partial
from itertools import repeat
import time
import uproot
import pickle
import sys
import math
import copy
import json
import cloudpickle
from tqdm.auto import tqdm
from collections import defaultdict
from cachetools import LRUCache
import lz4.frame as lz4f
from .processor import ProcessorABC
from .accumulator import (
    AccumulatorABC,
    value_accumulator,
    set_accumulator,
    dict_accumulator,
)
from .dataframe import (
    LazyDataFrame,
)
from ..nanoaod import NanoEvents
from ..util import _hash

try:
    from collections.abc import Mapping, Sequence
except ImportError:
    from collections import Mapping, Sequence


_PICKLE_PROTOCOL = pickle.HIGHEST_PROTOCOL
DEFAULT_METADATA_CACHE = LRUCache(100000)


# instrument xrootd source
if not hasattr(uproot.source.xrootd.XRootDSource, '_read_real'):
    def _read(self, chunkindex):
        self.bytesread = getattr(self, 'bytesread', 0) + self._chunkbytes
        return self._read_real(chunkindex)

    uproot.source.xrootd.XRootDSource._read_real = uproot.source.xrootd.XRootDSource._read
    uproot.source.xrootd.XRootDSource._read = _read


class FileMeta(object):
    __slots__ = ['dataset', 'filename', 'treename', 'metadata']

    def __init__(self, dataset, filename, treename, metadata=None):
        self.dataset = dataset
        self.filename = filename
        self.treename = treename
        self.metadata = metadata

    def __hash__(self):
        # As used to lookup metadata, no need for dataset
        return _hash((self.filename, self.treename))

    def __eq__(self, other):
        # In case of hash collisions
        return self.filename == other.filename and self.treename == other.treename

    def maybe_populate(self, cache):
        if cache and self in cache:
            self.metadata = cache[self]

    def populated(self, clusters=False):
        '''Return true if metadata is populated

        By default, only require bare minimum metadata (numentries, uuid)
        If clusters is True, then require cluster metadata to be populated
        '''
        if self.metadata is None:
            return False
        elif clusters and 'clusters' not in self.metadata:
            return False
        return True

    def chunks(self, target_chunksize, align_clusters):
        if not self.populated(clusters=align_clusters):
            raise RuntimeError
        if align_clusters:
            chunks = [0]
            for c in self.metadata['clusters']:
                if c >= chunks[-1] + target_chunksize:
                    chunks.append(c)
            if self.metadata['clusters'][-1] != chunks[-1]:
                chunks.append(self.metadata['clusters'][-1])
            for start, stop in zip(chunks[:-1], chunks[1:]):
                yield WorkItem(self.dataset, self.filename, self.treename, start, stop, self.metadata['uuid'])
        else:
            n = max(round(self.metadata['numentries'] / target_chunksize), 1)
            actual_chunksize = math.ceil(self.metadata['numentries'] / n)
            for index in range(n):
                start, stop = actual_chunksize * index, min(self.metadata['numentries'], actual_chunksize * (index + 1))
                yield WorkItem(self.dataset, self.filename, self.treename, start, stop, self.metadata['uuid'])


class WorkItem(object):
    __slots__ = ['dataset', 'filename', 'treename', 'entrystart', 'entrystop', 'fileuuid']

    def __init__(self, dataset, filename, treename, entrystart, entrystop, fileuuid):
        self.dataset = dataset
        self.filename = filename
        self.treename = treename
        self.entrystart = entrystart
        self.entrystop = entrystop
        self.fileuuid = fileuuid


class _compression_wrapper(object):
    def __init__(self, level, function, name=None):
        self.level = level
        self.function = function
        self.name = name

    def __str__(self):
        if self.name is not None:
            return self.name
        try:
            name = self.function.__name__
            if name == "<lambda>":
                return "lambda"
            return name
        except AttributeError:
            return str(self.function)

    # no @wraps due to pickle
    def __call__(self, *args, **kwargs):
        out = self.function(*args, **kwargs)
        return lz4f.compress(pickle.dumps(out, protocol=_PICKLE_PROTOCOL), compression_level=self.level)


def _maybe_decompress(item):
    if isinstance(item, AccumulatorABC):
        return item
    try:
        item = pickle.loads(lz4f.decompress(item))
        if isinstance(item, AccumulatorABC):
            return item
        raise RuntimeError
    except (RuntimeError, pickle.UnpicklingError):
        raise ValueError("Executors can only reduce accumulators or LZ4-compressed pickled accumulators")


def _iadd(output, result):
    output += _maybe_decompress(result)


class _reduce(object):
    def __init__(self):
        pass

    def __str__(self):
        return "reduce"

    def __call__(self, items):
        if len(items) == 0:
            raise ValueError("Empty list provided to reduction")
        out = items.pop()
        if isinstance(out, AccumulatorABC):
            # if dask has a cached result, we cannot alter it, so make a copy
            out = copy.deepcopy(out)
        else:
            out = _maybe_decompress(out)
        while items:
            out += _maybe_decompress(items.pop())
        return out


def _cancel(job):
    try:
        # this is not implemented with parsl AppFutures
        job.cancel()
    except NotImplementedError:
        pass


def _futures_handler(futures_set, output, status, unit, desc, add_fn, tailtimeout):
    start = time.time()
    last_job = start
    try:
        with tqdm(disable=not status, unit=unit, total=len(futures_set), desc=desc) as pbar:
            while len(futures_set) > 0:
                finished = set(job for job in futures_set if job.done())
                futures_set.difference_update(finished)
                while finished:
                    add_fn(output, finished.pop().result())
                    pbar.update(1)
                    last_job = time.time()
                time.sleep(0.5)
                if tailtimeout is not None and (time.time() - last_job) > tailtimeout and (last_job - start) > 0:
                    njobs = len(futures_set)
                    for job in futures_set:
                        _cancel(job)
                        pbar.update(1)
                    import warnings
                    warnings.warn('Stopped {} jobs early due to tailtimeout = {}'.format(njobs, tailtimeout))
                    break
    except KeyboardInterrupt:
        for job in futures_set:
            _cancel(job)
        if status:
            print("Received SIGINT, killed pending jobs.  Running jobs will continue to completion.", file=sys.stderr)
            print("Running jobs:", sum(1 for j in futures_set if j.running()), file=sys.stderr)
    except Exception:
        for job in futures_set:
            _cancel(job)
        raise


def iterative_executor(items, function, accumulator, **kwargs):
    """Execute in one thread iteratively

    Parameters
    ----------
        items : list
            List of input arguments
        function : callable
            A function to be called on each input, which returns an accumulator instance
        accumulator : AccumulatorABC
            An accumulator to collect the output of the function
        status : bool
            If true (default), enable progress bar
        unit : str
            Label of progress bar unit
        desc : str
            Label of progress bar description
        compression : int, optional
            Compress accumulator outputs in flight with LZ4, at level specified (default 1)
            Set to ``None`` for no compression.
    """
    if len(items) == 0:
        return accumulator
    status = kwargs.pop('status', True)
    unit = kwargs.pop('unit', 'items')
    desc = kwargs.pop('desc', 'Processing')
    clevel = kwargs.pop('compression', 1)
    if clevel is not None:
        function = _compression_wrapper(clevel, function)
    add_fn = _iadd
    for i, item in tqdm(enumerate(items), disable=not status, unit=unit, total=len(items), desc=desc):
        add_fn(accumulator, function(item))
    return accumulator


def futures_executor(items, function, accumulator, **kwargs):
    """Execute using multiple local cores using python futures

    Parameters
    ----------
        items : list
            List of input arguments
        function : callable
            A function to be called on each input, which returns an accumulator instance
        accumulator : AccumulatorABC
            An accumulator to collect the output of the function
        pool : concurrent.futures.Executor class or instance, optional
            The type of futures executor to use, defaults to ProcessPoolExecutor.
            You can pass an instance instead of a class to re-use an executor
        workers : int, optional
            Number of parallel processes for futures (default 1)
        status : bool, optional
            If true (default), enable progress bar
        unit : str, optional
            Label of progress bar unit (default: 'Processing')
        desc : str, optional
            Label of progress bar description (default: 'items')
        compression : int, optional
            Compress accumulator outputs in flight with LZ4, at level specified (default 1)
            Set to ``None`` for no compression.
        tailtimeout : int, optional
            Timeout requirement on job tails. Cancel all remaining jobs if none have finished
            in the timeout window.
    """
    if len(items) == 0:
        return accumulator
    pool = kwargs.pop('pool', concurrent.futures.ProcessPoolExecutor)
    workers = kwargs.pop('workers', 1)
    status = kwargs.pop('status', True)
    unit = kwargs.pop('unit', 'items')
    desc = kwargs.pop('desc', 'Processing')
    clevel = kwargs.pop('compression', 1)
    tailtimeout = kwargs.pop('tailtimeout', None)
    if clevel is not None:
        function = _compression_wrapper(clevel, function)
    add_fn = _iadd
    if isinstance(pool, concurrent.futures.Executor):
        futures = set(pool.submit(function, item) for item in items)
        _futures_handler(futures, accumulator, status, unit, desc, add_fn, tailtimeout)
    else:
        # assume its a class then
        with pool(max_workers=workers) as executor:
            futures = set(executor.submit(function, item) for item in items)
            _futures_handler(futures, accumulator, status, unit, desc, add_fn, tailtimeout)
    return accumulator


def dask_executor(items, function, accumulator, **kwargs):
    """Execute using dask futures

    Parameters
    ----------
        items : list
            List of input arguments
        function : callable
            A function to be called on each input, which returns an accumulator instance
        accumulator : AccumulatorABC
            An accumulator to collect the output of the function
        client : distributed.client.Client
            A dask distributed client instance
        treereduction : int, optional
            Tree reduction factor for output accumulators (default: 20)
        status : bool, optional
            If true (default), enable progress bar
        compression : int, optional
            Compress accumulator outputs in flight with LZ4, at level specified (default 1)
            Set to ``None`` for no compression.
        priority : int, optional
            Task priority, default 0
        retries : int, optional
            Number of retries for failed tasks (default: 3)
        heavy_input : serializable, optional
            Any value placed here will be broadcast to workers and joined to input
            items in a tuple (item, heavy_input) that is passed to function.
        function_name : str, optional
            Name of the function being passed

            .. note:: If ``heavy_input`` is set, ``function`` is assumed to be pure.
    """
    from dask.delayed import delayed
    if len(items) == 0:
        return accumulator
    client = kwargs.pop('client')
    ntree = kwargs.pop('treereduction', 20)
    status = kwargs.pop('status', True)
    clevel = kwargs.pop('compression', 1)
    priority = kwargs.pop('priority', 0)
    retries = kwargs.pop('retries', 3)
    heavy_input = kwargs.pop('heavy_input', None)
    function_name = kwargs.pop('function_name', None)
    reducer = _reduce()
    # secret options
    direct_heavy = kwargs.pop('direct_heavy', None)
    worker_affinity = kwargs.pop('worker_affinity', False)

    if clevel is not None:
        function = _compression_wrapper(clevel, function, name=function_name)
        reducer = _compression_wrapper(clevel, reducer)

    if heavy_input is not None:
        heavy_token = client.scatter(heavy_input, broadcast=True, hash=False, direct=direct_heavy)
        items = list(zip(items, repeat(heavy_token)))

    work = []
    if worker_affinity:
        workers = list(client.run(lambda: 0))

        def belongsto(workerindex, item):
            if heavy_input is not None:
                item = item[0]
            hashed = _hash((item.fileuuid, item.treename, item.entrystart, item.entrystop))
            return hashed % len(workers) == workerindex

        for workerindex, worker in enumerate(workers):
            work.extend(client.map(
                function,
                [item for item in items if belongsto(workerindex, item)],
                pure=(heavy_input is not None),
                priority=priority,
                retries=retries,
                workers={worker},
                allow_other_workers=False,
            ))
    else:
        work = client.map(
            function,
            items,
            pure=(heavy_input is not None),
            priority=priority,
            retries=retries,
        )
    while len(work) > 1:
        work = client.map(
            reducer,
            [work[i:i + ntree] for i in range(0, len(work), ntree)],
            pure=True,
            priority=priority,
            retries=retries,
        )
    work = work[0]
    if status:
        from distributed import progress
        # FIXME: fancy widget doesn't appear, have to live with boring pbar
        progress(work, multi=True, notebook=False)
    accumulator += _maybe_decompress(work.result())
    return accumulator


def parsl_executor(items, function, accumulator, **kwargs):
    """Execute using parsl pyapp wrapper

    Parameters
    ----------
        items : list
            List of input arguments
        function : callable
            A function to be called on each input, which returns an accumulator instance
        accumulator : AccumulatorABC
            An accumulator to collect the output of the function
        config : parsl.config.Config, optional
            A parsl DataFlow configuration object. Necessary if there is no active kernel

            .. note:: In general, it is safer to construct the DFK with ``parsl.load(config)`` prior to calling this function
        status : bool
            If true (default), enable progress bar
        unit : str
            Label of progress bar unit
        desc : str
            Label of progress bar description
        compression : int, optional
            Compress accumulator outputs in flight with LZ4, at level specified (default 1)
            Set to ``None`` for no compression.
        tailtimeout : int, optional
            Timeout requirement on job tails. Cancel all remaining jobs if none have finished
            in the timeout window.
    """
    if len(items) == 0:
        return accumulator
    import parsl
    from parsl.app.app import python_app
    from .parsl.timeout import timeout
    status = kwargs.pop('status', True)
    unit = kwargs.pop('unit', 'items')
    desc = kwargs.pop('desc', 'Processing')
    clevel = kwargs.pop('compression', 1)
    tailtimeout = kwargs.pop('tailtimeout', None)
    if clevel is not None:
        function = _compression_wrapper(clevel, function)
    add_fn = _iadd

    cleanup = False
    config = kwargs.pop('config', None)
    try:
        parsl.dfk()
    except RuntimeError:
        cleanup = True
        pass
    if cleanup and config is None:
        raise RuntimeError("No active parsl DataFlowKernel, must specify a config to construct one")
    elif not cleanup and config is not None:
        raise RuntimeError("An active parsl DataFlowKernel already exists")
    elif config is not None:
        parsl.clear()
        parsl.load(config)

    app = timeout(python_app(function))

    futures = set(app(item) for item in items)
    _futures_handler(futures, accumulator, status, unit, desc, add_fn, tailtimeout)

    if cleanup:
        parsl.dfk().cleanup()
        parsl.clear()

    return accumulator


def _work_function(item, processor_instance, flatten=False, savemetrics=False,
                   mmap=False, nano=False, cachestrategy=None, skipbadfiles=False,
                   retries=0, xrootdtimeout=None):
    if processor_instance == 'heavy':
        item, processor_instance = item
    if not isinstance(processor_instance, ProcessorABC):
        processor_instance = cloudpickle.loads(lz4f.decompress(processor_instance))
    if mmap:
        localsource = {}
    else:
        opts = dict(uproot.FileSource.defaults)
        opts.update({'parallel': None})

        def localsource(path):
            return uproot.FileSource(path, **opts)

    import warnings
    out = processor_instance.accumulator.identity()
    retry_count = 0
    while retry_count <= retries:
        try:
            from uproot.source.xrootd import XRootDSource
            xrootdsource = XRootDSource.defaults
            xrootdsource['timeout'] = xrootdtimeout
            file = uproot.open(item.filename, localsource=localsource, xrootdsource=xrootdsource)
            if nano:
                cache = None
                if cachestrategy == 'dask-worker':
                    from distributed import get_worker
                    from .dask import ColumnCache
                    worker = get_worker()
                    try:
                        cache = worker.plugins[ColumnCache.name]
                    except KeyError:
                        # emit warning if not found?
                        pass
                df = NanoEvents.from_file(
                    file=file,
                    treename=item.treename,
                    entrystart=item.entrystart,
                    entrystop=item.entrystop,
                    metadata={'dataset': item.dataset},
                    cache=cache,
                )
            else:
                tree = file[item.treename]
                df = LazyDataFrame(tree, item.entrystart, item.entrystop, flatten=flatten)
                df['dataset'] = item.dataset
            tic = time.time()
            out = processor_instance.process(df)
            toc = time.time()
            metrics = dict_accumulator()
            if savemetrics:
                if isinstance(file.source, uproot.source.xrootd.XRootDSource):
                    metrics['bytesread'] = value_accumulator(int, file.source.bytesread)
                    metrics['dataservers'] = set_accumulator({file.source._source.get_property('DataServer')})
                metrics['columns'] = set_accumulator(df.materialized)
                metrics['entries'] = value_accumulator(int, df.size)
                metrics['processtime'] = value_accumulator(float, toc - tic)
            wrapped_out = dict_accumulator({'out': out, 'metrics': metrics})
            file.source.close()
            break
        # catch xrootd errors and optionally skip
        # or retry to read the file
        except OSError as e:
            if not skipbadfiles:
                raise e
            else:
                w_str = 'Bad file source %s.' % item.filename
                if retries:
                    w_str += ' Attempt %d of %d.' % (retry_count + 1, retries + 1)
                    if retry_count + 1 < retries:
                        w_str += ' Will retry.'
                    else:
                        w_str += ' Skipping.'
                else:
                    w_str += ' Skipping.'
                warnings.warn(w_str)
            metrics = dict_accumulator()
            if savemetrics:
                metrics['bytesread'] = value_accumulator(int, 0)
                metrics['dataservers'] = set_accumulator({})
                metrics['columns'] = set_accumulator({})
                metrics['entries'] = value_accumulator(int, 0)
                metrics['processtime'] = value_accumulator(float, 0)
            wrapped_out = dict_accumulator({'out': out, 'metrics': metrics})
        except Exception as e:
            if retries == retry_count:
                raise e
            w_str = 'Attempt %d of %d. Will retry.' % (retry_count + 1, retries + 1)
            warnings.warn(w_str)
        retry_count += 1

    return wrapped_out


def _normalize_fileset(fileset, treename):
    if isinstance(fileset, str):
        with open(fileset) as fin:
            fileset = json.load(fin)
    for dataset, filelist in fileset.items():
        if isinstance(filelist, dict):
            local_treename = filelist['treename'] if 'treename' in filelist else treename
            filelist = filelist['files']
        elif isinstance(filelist, list):
            if treename is None:
                raise ValueError('treename must be specified if the fileset does not contain tree names')
            local_treename = treename
        else:
            raise ValueError('list of filenames in fileset must be a list or a dict')
        for filename in filelist:
            yield FileMeta(dataset, filename, local_treename)


def _get_metadata(item, skipbadfiles=False, retries=0, xrootdtimeout=None, align_clusters=False):
    import warnings
    out = set_accumulator()
    retry_count = 0
    while retry_count <= retries:
        try:
            # add timeout option according to modified uproot numentries defaults
            xrootdsource = {"timeout": xrootdtimeout, "chunkbytes": 32 * 1024, "limitbytes": 1024**2, "parallel": False}
            file = uproot.open(item.filename, xrootdsource=xrootdsource)
            tree = file[item.treename]
            metadata = {'numentries': tree.numentries, 'uuid': file._context.uuid}
            if align_clusters:
                metadata['clusters'] = [0] + list(c[1] for c in tree.clusters())
            out = set_accumulator([FileMeta(item.dataset, item.filename, item.treename, metadata)])
            break
        except OSError as e:
            if not skipbadfiles:
                raise e
            else:
                w_str = 'Bad file source %s.' % item.filename
                if retries:
                    w_str += ' Attempt %d of %d.' % (retry_count + 1, retries + 1)
                    if retry_count + 1 < retries:
                        w_str += ' Will retry.'
                    else:
                        w_str += ' Skipping.'
                else:
                    w_str += ' Skipping.'
                warnings.warn(w_str)
        except Exception as e:
            if retries == retry_count:
                raise e
            w_str = 'Attempt %d of %d. Will retry.' % (retry_count + 1, retries + 1)
            warnings.warn(w_str)
        retry_count += 1
    return out


def run_uproot_job(fileset,
                   treename,
                   processor_instance,
                   executor,
                   executor_args={},
                   pre_executor=None,
                   pre_args=None,
                   chunksize=100000,
                   maxchunks=None,
                   metadata_cache=None,
                   ):
    '''A tool to run a processor using uproot for data delivery

    A convenience wrapper to submit jobs for a file set, which is a
    dictionary of dataset: [file list] entries.  Supports only uproot
    reading, via the LazyDataFrame class.  For more customized processing,
    e.g. to read other objects from the files and pass them into data frames,
    one can write a similar function in their user code.

    Parameters
    ----------
        fileset : dict
            A dictionary ``{dataset: [file, file], }``
            Optionally, if some files' tree name differ, the dictionary can be specified:
            ``{dataset: {'treename': 'name', 'files': [file, file]}, }``
        treename : str
            name of tree inside each root file, can be ``None``;
            treename can also be defined in fileset, which will override the passed treename
        processor_instance : ProcessorABC
            An instance of a class deriving from ProcessorABC
        executor : callable
            A function that takes 3 arguments: items, function, accumulator
            and performs some action equivalent to:
            ``for item in items: accumulator += function(item)``
        executor_args : dict, optional
            Arguments to pass to executor.  See `iterative_executor`,
            `futures_executor`, `dask_executor`, or `parsl_executor` for available options.
            Some options that affect the behavior of this function:
            'savemetrics' saves some detailed metrics for xrootd processing (default False);
            'flatten' removes any jagged structure from the input files (default False);
            'nano' builds the dataframe as a `NanoEvents` object rather than `LazyDataFrame` (default False);
            'processor_compression' sets the compression level used to send processor instance to workers (default 1).
            'skipbadfiles' instead of failing on a bad file, skip it (default False)
            'retries' optionally retry n times (default 0)
            'xrootdtimeout' timeout for xrootd read (seconds)
            'tailtimeout' timeout requirement on job tails (seconds)
            'align_clusters' aligns the chunks to natural boundaries in the ROOT files
        pre_executor : callable
            A function like executor, used to calculate fileset metadata
            Defaults to executor
        pre_args : dict, optional
            Similar to executor_args, defaults to executor_args
        chunksize : int, optional
            Maximum number of entries to process at a time in the data frame, default: 100k
        maxchunks : int, optional
            Maximum number of chunks to process per dataset
            Defaults to processing the whole dataset
        metadata_cache : mapping, optional
            A dict-like object to use as a cache for (file, tree) metadata that is used to
            determine chunking.  Defaults to a in-memory LRU cache that holds 100k entries
            (about 1MB depending on the length of filenames, etc.)  If you edit an input file
            (please don't) during a session, the session can be restarted to clear the cache.
    '''
    if not isinstance(fileset, (Mapping, str)):
        raise ValueError("Expected fileset to be a mapping dataset: list(files) or filename")
    if not isinstance(processor_instance, ProcessorABC):
        raise ValueError("Expected processor_instance to derive from ProcessorABC")

    if pre_executor is None:
        pre_executor = executor
    if pre_args is None:
        pre_args = dict(executor_args)
    if metadata_cache is None:
        metadata_cache = DEFAULT_METADATA_CACHE

    fileset = list(_normalize_fileset(fileset, treename))
    for filemeta in fileset:
        filemeta.maybe_populate(metadata_cache)

    # pop _get_metdata args here (also sent to _work_function)
    skipbadfiles = executor_args.pop('skipbadfiles', False)
    retries = executor_args.pop('retries', 0)
    xrootdtimeout = executor_args.pop('xrootdtimeout', None)
    align_clusters = executor_args.pop('align_clusters', False)
    metadata_fetcher = partial(_get_metadata,
                               skipbadfiles=skipbadfiles,
                               retries=retries,
                               xrootdtimeout=xrootdtimeout,
                               align_clusters=align_clusters,
                               )

    chunks = []
    if maxchunks is None:
        # this is a bit of an abuse of map-reduce but ok
        to_get = set(filemeta for filemeta in fileset if not filemeta.populated(clusters=align_clusters))
        if len(to_get) > 0:
            out = set_accumulator()
            pre_arg_override = {
                'desc': 'Preprocessing',
                'unit': 'file',
                'compression': None,
                'tailtimeout': None,
                'worker_affinity': False,
            }
            pre_args.update(pre_arg_override)
            executor(to_get, metadata_fetcher, out, **pre_args)
            while out:
                item = out.pop()
                metadata_cache[item] = item.metadata
            for filemeta in fileset:
                filemeta.maybe_populate(metadata_cache)
        while fileset:
            filemeta = fileset.pop()
            if skipbadfiles and not filemeta.populated(clusters=align_clusters):
                continue
            for chunk in filemeta.chunks(chunksize, align_clusters):
                chunks.append(chunk)
    else:
        # get just enough file info to compute chunking
        nchunks = defaultdict(int)
        while fileset:
            filemeta = fileset.pop()
            if nchunks[filemeta.dataset] >= maxchunks:
                continue
            if not filemeta.populated(clusters=align_clusters):
                filemeta.metadata = metadata_fetcher(filemeta).pop().metadata
                metadata_cache[filemeta] = filemeta.metadata
            if skipbadfiles and not filemeta.populated(clusters=align_clusters):
                continue
            for chunk in filemeta.chunks(chunksize, align_clusters):
                chunks.append(chunk)
                nchunks[filemeta.dataset] += 1
                if nchunks[filemeta.dataset] >= maxchunks:
                    break

    # pop all _work_function args here
    savemetrics = executor_args.pop('savemetrics', False)
    flatten = executor_args.pop('flatten', False)
    mmap = executor_args.pop('mmap', False)
    nano = executor_args.pop('nano', False)
    cachestrategy = executor_args.pop('cachestrategy', None)
    pi_compression = executor_args.pop('processor_compression', 1)
    if pi_compression is None:
        pi_to_send = processor_instance
    else:
        pi_to_send = lz4f.compress(cloudpickle.dumps(processor_instance), compression_level=pi_compression)
    closure = partial(
        _work_function,
        flatten=flatten,
        savemetrics=savemetrics,
        mmap=mmap,
        nano=nano,
        cachestrategy=cachestrategy,
        skipbadfiles=skipbadfiles,
        retries=retries,
        xrootdtimeout=xrootdtimeout,
    )
    # hack around dask/dask#5503 which is really a silly request but here we are
    if executor is dask_executor:
        executor_args['heavy_input'] = pi_to_send
        closure = partial(closure, processor_instance='heavy')
    else:
        closure = partial(closure, processor_instance=pi_to_send)

    out = processor_instance.accumulator.identity()
    wrapped_out = dict_accumulator({'out': out, 'metrics': dict_accumulator()})
    exe_args = {
        'unit': 'chunk',
        'function_name': type(processor_instance).__name__,
    }
    exe_args.update(executor_args)
    executor(chunks, closure, wrapped_out, **exe_args)
    wrapped_out['metrics']['chunks'] = value_accumulator(int, len(chunks))
    processor_instance.postprocess(out)
    if savemetrics:
        return out, wrapped_out['metrics']
    return out


def run_parsl_job(fileset, treename, processor_instance, executor, executor_args={}, chunksize=200000):
    '''A wrapper to submit parsl jobs

    .. note:: Deprecated in favor of `run_uproot_job` with the `parsl_executor`

    Jobs are specified by a file set, which is a dictionary of
    dataset: [file list] entries.  Supports only uproot reading,
    via the LazyDataFrame class.  For more customized processing,
    e.g. to read other objects from the files and pass them into data frames,
    one can write a similar function in their user code.

    Parameters
    ----------
        fileset : dict
            dictionary {dataset: [file, file], }
        treename : str
            name of tree inside each root file
        processor_instance : ProcessorABC
            An instance of a class deriving from ProcessorABC
        executor : coffea.processor.parsl.parsl_executor
            Must be the parsl executor, or otherwise derive from
            ``coffea.processor.parsl.ParslExecutor``
        executor_args : dict
            Extra arguments to pass to executor.  Special options
            interpreted here: 'config' provides a parsl dataflow
            configuration.
        chunksize : int, optional
            Number of entries to process at a time in the data frame

    '''

    try:
        import parsl
    except ImportError as e:
        print('you must have parsl installed to call run_parsl_job()!', file=sys.stderr)
        raise e

    import warnings

    warnings.warn("run_parsl_job is deprecated and will be removed in 0.7.0, replaced by run_uproot_job",
                  DeprecationWarning)

    from .parsl.parsl_executor import ParslExecutor
    from .parsl.detail import _default_cfg

    if not isinstance(fileset, Mapping):
        raise ValueError("Expected fileset to be a mapping dataset: list(files)")
    if not isinstance(processor_instance, ProcessorABC):
        raise ValueError("Expected processor_instance to derive from ProcessorABC")
    if isinstance(executor, ParslExecutor):
        warnings.warn("ParslExecutor class is deprecated replacing with processor.parsl_executor",
                      DeprecationWarning)
        executor = parsl_executor
    elif executor == parsl_executor:
        pass
    else:
        raise ValueError("Expected executor to derive from ParslExecutor or be executor.parsl_executor")

    executor_args.setdefault('config', _default_cfg)
    executor_args.setdefault('timeout', 180)
    executor_args.setdefault('chunking_timeout', 10)
    executor_args.setdefault('flatten', False)
    executor_args.setdefault('compression', 0)
    executor_args.setdefault('skipbadfiles', False)
    executor_args.setdefault('retries', 0)
    executor_args.setdefault('xrootdtimeout', None)

    try:
        parsl.dfk()
        executor_args.pop('config')
    except RuntimeError:
        pass

    output = run_uproot_job(fileset,
                            treename,
                            processor_instance=processor_instance,
                            executor=executor,
                            executor_args=executor_args)

    return output


def run_spark_job(fileset, processor_instance, executor, executor_args={},
                  spark=None, partitionsize=200000, thread_workers=16):
    '''A wrapper to submit spark jobs

    A convenience wrapper to submit jobs for spark datasets, which is a
    dictionary of dataset: [file list] entries.  Presently supports reading of
    parquet files converted from root.  For more customized processing,
    e.g. to read other objects from the files and pass them into data frames,
    one can write a similar function in their user code.

    Parameters
    ----------
        fileset : dict
            dictionary {dataset: [file, file], }
        processor_instance : ProcessorABC
            An instance of a class deriving from ProcessorABC

            .. note:: The processor instance must define all the columns in data and MC that it reads as ``.columns``
        executor:
            anything that inherits from `SparkExecutor` like `spark_executor`

            In general, a function that takes 3 arguments: items, function accumulator
            and performs some action equivalent to:
            for item in items: accumulator += function(item)
        executor_args:
            arguments to send to the creation of a spark session
        spark:
            an optional already created spark instance

            if ``None`` then we create an ephemeral spark instance using a config
        partitionsize:
            partition size to try to aim for (coalescese only, repartition too expensive)
        thread_workers:
            how many spark jobs to let fly in parallel during processing steps
    '''

    try:
        import pyspark
    except ImportError as e:
        print('you must have pyspark installed to call run_spark_job()!', file=sys.stderr)
        raise e

    from packaging import version
    import pyarrow as pa
    import warnings
    arrow_env = ('ARROW_PRE_0_15_IPC_FORMAT', '1')
    if (version.parse(pa.__version__) >= version.parse('0.15.0') and
        version.parse(pyspark.__version__) < version.parse('3.0.0')):
        import os
        if (arrow_env[0] not in os.environ or
            os.environ[arrow_env[0]] != arrow_env[1]):
            warnings.warn('If you are using pyarrow >= 0.15.0, make sure to set %s=%s in your environment!' % arrow_env)

    import pyspark.sql
    from .spark.spark_executor import SparkExecutor
    from .spark.detail import _spark_initialize, _spark_stop, _spark_make_dfs

    if not isinstance(fileset, Mapping):
        raise ValueError("Expected fileset to be a mapping dataset: list(files)")
    if not isinstance(processor_instance, ProcessorABC):
        raise ValueError("Expected processor_instance to derive from ProcessorABC")
    if not isinstance(executor, SparkExecutor):
        raise ValueError("Expected executor to derive from SparkExecutor")

    executor_args.setdefault('config', None)
    executor_args.setdefault('file_type', 'parquet')
    executor_args.setdefault('laurelin_version', '0.3.0')
    executor_args.setdefault('treeName', 'Events')
    executor_args.setdefault('flatten', False)
    executor_args.setdefault('cache', True)
    executor_args.setdefault('skipbadfiles', False)
    executor_args.setdefault('retries', 0)
    executor_args.setdefault('xrootdtimeout', None)
    file_type = executor_args['file_type']
    treeName = executor_args['treeName']
    flatten = executor_args['flatten']
    use_cache = executor_args['cache']

    if executor_args['config'] is None:
        executor_args.pop('config')

    # initialize spark if we need to
    # if we initialize, then we deconstruct
    # when we're done
    killSpark = False
    if spark is None:
        spark = _spark_initialize(**executor_args)
        killSpark = True
        use_cache = False  # if we always kill spark then we cannot use the cache
    else:
        if not isinstance(spark, pyspark.sql.session.SparkSession):
            raise ValueError("Expected 'spark' to be a pyspark.sql.session.SparkSession")

    dfslist = {}
    if executor._cacheddfs is None:
        dfslist = _spark_make_dfs(spark, fileset, partitionsize, processor_instance.columns,
                                  thread_workers, file_type, treeName)

    output = processor_instance.accumulator.identity()
    executor(spark, dfslist, processor_instance, output, thread_workers, use_cache, flatten)
    processor_instance.postprocess(output)

    if killSpark:
        _spark_stop(spark)
        del spark
        spark = None

    return output



from threading import Lock
import pandas
import numpy
import numba


class BTagScaleFactor:
    '''A class holding one complete BTag scale factor for a given working point

    Parameters
    ----------
        filename : str
            The BTag-formatted CSV file to read (accepts .csv, .csv.gz, etc.)
            See pandas read_csv for all supported compressions.
        workingpoint : str or int
            The working point, one of LOOSE, MEDIUM, TIGHT, or RESHAPE (0-3, respectively)
        methods : str, optional
            The scale factor derivation method to use for each flavor, 'b,c,light' respectively.
            Defaults to 'comb,comb,incl'
        keep_df : bool, optional
            If set true, keep the parsed dataframe as an attribute (.df) for later inspection
    '''
    LOOSE, MEDIUM, TIGHT, RESHAPE = range(4)
    FLAV_B, FLAV_C, FLAV_UDSG = range(3)
    _formulaLock = Lock()
    _formulaCache = {}
    _btvflavor = numpy.array([0, 1, 2, 3])
    _flavor = numpy.array([0, 4, 5, 6])
    _wpString = {'loose': LOOSE, 'medium': MEDIUM, 'tight': TIGHT}
    _expectedColumns = [
        'OperatingPoint', 'measurementType', 'sysType', 'jetFlavor', 'etaMin',
        'etaMax', 'ptMin', 'ptMax', 'discrMin', 'discrMax', 'formula'
    ]

    @classmethod
    def readcsv(cls, filename):
        '''Reads a BTag-formmated CSV file into a pandas dataframe

        This function also merges the bin min and max into a tuple representing the bin

        Parameters
        ----------
            filename : str
                The file to open

        Returns
        -------
            df : pandas.DataFrame
                A dataframe containing all info in the file
            discriminator : str
                The name of the discriminator the correction map is for
        '''
        df = pandas.read_csv(filename, skipinitialspace=True)
        discriminator = df.columns[0].split(';')[0]

        def cleanup(colname):
            if ';' in colname:
                _, colname = colname.split(';')
            return colname.strip()
        df.rename(columns=cleanup, inplace=True)
        if not list(df.columns) == BTagScaleFactor._expectedColumns:
            raise RuntimeError('Columns in BTag scale factor file %s as expected' % filename)
        for var in ['eta', 'pt', 'discr']:
            df[var + 'Bin'] = list(zip(df[var + 'Min'], df[var + 'Max']))
            del df[var + 'Min']
            del df[var + 'Max']
        return df, discriminator

    def __init__(self, filename, workingpoint, methods='comb,comb,incl', keep_df=False):
        if workingpoint not in [0, 1, 2, 3]:
            try:
                workingpoint = BTagScaleFactor._wpString[workingpoint.lower()]
            except (KeyError, AttributeError):
                raise ValueError('Unrecognized working point')
        methods = methods.split(',')
        self.workingpoint = workingpoint
        df, self.discriminator = BTagScaleFactor.readcsv(filename)
        cut = (df['jetFlavor'] == self.FLAV_B) & (df['measurementType'] == methods[0])
        if len(methods) > 1:
            cut |= (df['jetFlavor'] == self.FLAV_C) & (df['measurementType'] == methods[1])
        if len(methods) > 2:
            cut |= (df['jetFlavor'] == self.FLAV_UDSG) & (df['measurementType'] == methods[2])
        cut &= df['OperatingPoint'] == workingpoint
        df = df[cut]
        mavailable = list(df['measurementType'].unique())
        if not all(m in mavailable for m in methods):
            raise ValueError('Unrecognized jet correction method, available: %r' % mavailable)
        df = df.set_index(['sysType', 'jetFlavor', 'etaBin', 'ptBin', 'discrBin']).sort_index()
        if keep_df:
            self.df = df
        self._corrections = {}
        for syst in list(df.index.levels[0]):
            corr = df.loc[syst]
            allbins = list(corr.index)
            # NOTE: here we force the class to assume abs(eta) based on lack of examples where signed eta is used
            edges_eta = numpy.array(sorted(set(abs(x) for tup in corr.index.levels[1] for x in tup)))
            edges_pt = numpy.array(sorted(set(x for tup in corr.index.levels[2] for x in tup)))
            edges_discr = numpy.array(sorted(set(x for tup in corr.index.levels[3] for x in tup)))
            alledges = numpy.meshgrid(self._btvflavor[:-1], edges_eta[:-1], edges_pt[:-1], edges_discr[:-1], indexing='ij')
            mapping = numpy.full(alledges[0].shape, -1)

            def findbin(btvflavor, eta, pt, discr):
                for i, (fbin, ebin, pbin, dbin) in enumerate(allbins):
                    if btvflavor == fbin and ebin[0] <= eta < ebin[1] and pbin[0] <= pt < pbin[1] and dbin[0] <= discr < dbin[1]:
                        return i
                return -1

            for idx, _ in numpy.ndenumerate(mapping):
                btvflavor, eta, pt, discr = (x[idx] for x in alledges)
                mapping[idx] = findbin(btvflavor, eta, pt, discr)

            self._corrections[syst] = (
                edges_eta,
                edges_pt,
                edges_discr,
                mapping,
                numpy.array(corr['formula']),
            )
        self._compiled = {}

    def __getstate__(self):
        state = dict(self.__dict__)
        state.pop('_compiled')
        return state

    def __setstate__(self, state):
        self.__dict__ = state
        self._compiled = {}

    @classmethod
    def _compile(cls, formula):
        with BTagScaleFactor._formulaLock:
            try:
                return BTagScaleFactor._formulaCache[formula]
            except KeyError:
                if 'x' in formula:
                    feval = eval('lambda x: ' + formula, {'log': numpy.log, 'sqrt': numpy.sqrt})
                    out = numba.vectorize([
                        numba.float32(numba.float32),
                        numba.float64(numba.float64),
                    ])(feval)
                else:
                    val = eval(formula, {'log': numpy.log, 'sqrt': numpy.sqrt})

                    def duck(_, out, where):
                        out[where] = val
                    out = duck
                BTagScaleFactor._formulaCache[formula] = out
                return out

    def _lookup(self, axis, values):
        if len(axis) == 2:
            return numpy.zeros(shape=values.shape, dtype=numpy.uint)
        return numpy.clip(numpy.searchsorted(axis, values, side='right') - 1, 0, len(axis) - 2)

    def eval(self, systematic, flavor, abseta, pt, discr=None, ignore_missing=False):
        '''Evaluate this scale factor as a function of jet properties

        Parameters
        ----------
            systematic : str
                Which systematic to evaluate. Nominal correction is 'central', the options depend
                on the scale factor and method
            flavor : numpy.ndarray or awkward.Array
                The generated jet hadron flavor, following the enumeration:
                0: uds quark or gluon, 4: charm quark, 5: bottom quark
            abseta : numpy.ndarray or awkward.Array
                The absolute value of the jet pseudorapitiy
            pt : numpy.ndarray or awkward.Array
                The jet transverse momentum
            discr : numpy.ndarray or awkward.Array, optional
                The jet tagging discriminant value (default None), optional for all scale factors except
                the reshaping scale factor
            ignore_missing : bool, optional
                If set true, any values that have no correction will return 1. instead of throwing
                an exception. Out-of-bounds values are always clipped to the nearest bin.

        Returns
        -------
            out : numpy.ndarray or awkward.Array
                An array with shape matching ``pt``, containing the per-jet scale factor
        '''
        if self.workingpoint == BTagScaleFactor.RESHAPE and discr is None:
            raise ValueError('RESHAPE scale factor requires a discriminant array')
        try:
            functions = self._compiled[systematic]
        except KeyError:
            functions = [BTagScaleFactor._compile(f) for f in self._corrections[systematic][-1]]
            self._compiled[systematic] = functions
        try:
            flavor.counts
            jin, flavor = flavor, flavor.flatten()
            abseta = abseta.flatten()
            pt = pt.flatten()
            discr = discr.flatten() if discr is not None else None
        except AttributeError:
            jin = None
        corr = self._corrections[systematic]
        idx = (
            2 - self._lookup(self._flavor, flavor),  # transform to btv definiton
            self._lookup(corr[0], abseta),
            self._lookup(corr[1], pt),
            self._lookup(corr[2], discr) if discr is not None else 0,
        )
        mapidx = corr[3][idx]
        out = numpy.ones(mapidx.shape, dtype=pt.dtype)
        for ifunc in numpy.unique(mapidx):
            if ifunc < 0 and not ignore_missing:
                raise ValueError('No correction was available for some items')
            func = functions[ifunc]
            if self.workingpoint == BTagScaleFactor.RESHAPE:
                var = numpy.clip(discr, corr[2][0], corr[2][-1])
            else:
                var = numpy.clip(pt, corr[1][0], corr[1][-1])
            func(var, out=out, where=(mapidx == ifunc))

        if jin is not None:
            out = jin.copy(content=out)
        return out

    def __call__(self, systematic, flavor, abseta, pt, discr=None, ignore_missing=False):
        return self.eval(systematic, flavor, abseta, pt, discr, ignore_missing)



import pytest
from coffea.btag_tools import BTagScaleFactor
from dummy_distributions import dummy_jagged_eta_pt
import numpy

def test_BTagScalefactor():
    sf1 = BTagScaleFactor('tests/samples/testBTagSF.btag.csv', 'medium')
    sf2 = BTagScaleFactor('tests/samples/DeepCSV_102XSF_V1.btag.csv.gz', BTagScaleFactor.RESHAPE, 'iterativefit')
    sf3 = BTagScaleFactor('tests/samples/DeepCSV_102XSF_V1.btag.csv.gz', BTagScaleFactor.TIGHT)
    sf4 = BTagScaleFactor('tests/samples/DeepCSV_2016LegacySF_V1_TuneCP5.btag.csv.gz', BTagScaleFactor.RESHAPE, 'iterativefit', keep_df=True)
    # import pdb; pdb.set_trace()

    counts, test_eta, test_pt = dummy_jagged_eta_pt()
    test_flavor = numpy.random.choice([0, 4, 5], size=len(test_eta))
    test_allb = numpy.ones_like(test_flavor) * 5
    test_discr = numpy.random.rand(len(test_eta))

    sf1.eval('central', test_flavor, test_eta, test_pt, test_discr)
    sf1.eval('up', test_flavor, test_eta, test_pt)
    sf2.eval('central', test_allb, test_eta, test_pt, test_discr)
    with pytest.raises(ValueError):
        sf2.eval('up', test_allb, test_eta, test_pt)
    sf3.eval('central', test_flavor, test_eta, test_pt, test_discr)
    sf3.eval('up', test_flavor, test_eta, test_pt)
    with pytest.raises(ValueError):
        sf4.eval('central', test_flavor, test_eta, test_pt, test_discr)
    sf4.eval('central', test_allb, test_eta, test_pt, test_discr)



from __future__ import print_function, division

import os
from coffea import lookup_tools
import uproot
from coffea.util import awkward
from coffea.util import numpy as np
from awkward import JaggedArray
from coffea.nanoaod import NanoEvents

from dummy_distributions import dummy_jagged_eta_pt


def test_extractor_exceptions():
    extractor = lookup_tools.extractor()

    # test comments
    extractor.add_weight_sets(["#testSF2d asdfgh tests/samples/testSF2d.histo.root"])
    
    # test malformed statement
    try:
        extractor.add_weight_sets(["testSF2d testSF2d asdfgh tests/samples/testSF2d.histo.root"])
    except Exception as e:
        assert(e.args[0] == '"testSF2d testSF2d asdfgh tests/samples/testSF2d.histo.root" not formatted as "<local name> <name> <weights file>"')
    
    # test not existant file entry
    try:
         extractor.add_weight_sets(["testSF2d asdfgh tests/samples/testSF2d.histo.root"])
    except Exception as e:
        assert(e.args[0] == 'Weights named "asdfgh" not in tests/samples/testSF2d.histo.root!')

    # test unfinalized evaluator
    try:
        extractor.add_weight_sets(["testSF2d scalefactors_Tight_Electron tests/samples/testSF2d.histo.root"])
        extractor.make_evaluator()
    except Exception as e:
        assert(e.args[0] == 'Cannot make an evaluator from unfinalized extractor!')

def test_evaluator_exceptions():
    extractor = lookup_tools.extractor()
    extractor.add_weight_sets(["testSF2d scalefactors_Tight_Electron tests/samples/testSF2d.histo.root"])

    counts, test_eta, test_pt = dummy_jagged_eta_pt()
    test_eta_jagged = awkward.JaggedArray.fromcounts(counts, test_eta)
    test_pt_jagged = awkward.JaggedArray.fromcounts(counts, test_pt)

    extractor.finalize()
    evaluator = extractor.make_evaluator()
    
    try:
        test_out = evaluator["testSF2d"](test_pt_jagged, test_eta)
    except Exception as e:
        assert(e.args[0] == 'do not mix JaggedArrays and numpy arrays when calling a derived class of lookup_base')

def test_evaluator_exceptions():
    from coffea.lookup_tools.lookup_base import lookup_base
    try:
        lookup_base()._evaluate()
    except NotImplementedError:
        pass

def test_root_scalefactors():
    extractor = lookup_tools.extractor()
    extractor.add_weight_sets(["testSF2d scalefactors_Tight_Electron tests/samples/testSF2d.histo.root"])
    
    extractor.finalize(reduce_list=['testSF2d'])

    evaluator = extractor.make_evaluator()

    counts, test_eta, test_pt = dummy_jagged_eta_pt()
    
    # test flat eval
    test_out = evaluator["testSF2d"](test_eta, test_pt)

    # print it
    print(evaluator["testSF2d"])
    
    # test structured eval
    test_eta_jagged = awkward.JaggedArray.fromcounts(counts, test_eta)
    test_pt_jagged = awkward.JaggedArray.fromcounts(counts, test_pt)
    test_out_jagged = evaluator["testSF2d"](test_eta_jagged, test_pt_jagged)

    assert (test_out_jagged.counts==counts).all()
    assert (test_out==test_out_jagged.flatten()).all()

    # From make_expected_lookup.py
    expected_output = np.array([
       0.90780139, 0.82748538, 0.86332178, 0.86332178, 0.97981155,
       0.79701495, 0.88245934, 0.82857144, 0.91884059, 0.97466666,
       0.94072163, 1.00775194, 0.82748538, 1.00775194, 0.97203946,
       0.98199672, 0.80655736, 0.90893763, 0.88245934, 0.79701495,
       0.82748538, 0.82857144, 0.91884059, 0.90893763, 0.97520661,
       0.97520661, 0.82748538, 0.91884059, 0.97203946, 0.88245934,
       0.79701495, 0.9458763 , 1.00775194, 0.80655736, 1.00775194,
       1.00775194, 0.98976982, 0.98976982, 0.86332178, 0.94072163,
       0.80655736, 0.98976982, 0.96638656, 0.9458763 , 0.90893763,
       0.9529984 , 0.9458763 , 0.9529984 , 0.80655736, 0.80655736,
       0.80655736, 0.98976982, 0.97466666, 0.98199672, 0.86332178,
       1.03286386, 0.94072163, 1.03398061, 0.82857144, 0.80655736,
       1.00775194, 0.80655736])

    diff = np.abs(test_out-expected_output)
    print("Max diff: %.16f" % diff.max())
    print("Median diff: %.16f" % np.median(diff))
    print("Diff over threshold rate: %.1f %%" % (100*(diff >= 1.e-8).sum()/diff.size))
    assert (diff < 1.e-8).all()


def test_btag_csv_scalefactors():
    extractor = lookup_tools.extractor()
    extractor.add_weight_sets(["testBTag * tests/samples/testBTagSF.btag.csv",
                               "* * tests/samples/DeepCSV_102XSF_V1.btag.csv.gz",
                               "* * tests/samples/DeepJet_102XSF_WP_V1.btag.csv.gz",
                               ])
    extractor.finalize()

    evaluator = extractor.make_evaluator()

    counts, test_eta, test_pt = dummy_jagged_eta_pt()
    # discriminant used for reshaping, zero otherwise
    test_discr = np.zeros_like(test_eta)

    print(evaluator['testBTagCSVv2_1_comb_up_0'])
    
    print(evaluator['DeepCSV_1_comb_up_0'])
    
    print(evaluator['btagsf_1_comb_up_0'])
    
    sf_out = evaluator['testBTagCSVv2_1_comb_up_0'](test_eta, test_pt, test_discr)


def test_histo_json_scalefactors():
    extractor = lookup_tools.extractor()
    extractor.add_weight_sets(["testJson * tests/samples/EIDISO_WH_out.histo.json"])
    extractor.finalize()
    
    evaluator = extractor.make_evaluator()
    
    counts, test_eta, test_pt = dummy_jagged_eta_pt()
    
    sf_out = evaluator['testJsonEIDISO_WH/eta_pt_ratio_value'](test_eta, test_pt)
    sf_err_out = evaluator['testJsonEIDISO_WH/eta_pt_ratio_error'](test_eta, test_pt)
    print(sf_out)
    print(sf_err_out)

def test_jec_txt_scalefactors():
    extractor = lookup_tools.extractor()
    extractor.add_weight_sets([
        "testJEC * tests/samples/Fall17_17Nov2017_V32_MC_L2Relative_AK4PFPuppi.jec.txt",
        "* * tests/samples/Summer16_07Aug2017_V11_L1fix_MC_L2Relative_AK4PFchs.jec.txt.gz",
        "* * tests/samples/Fall17_17Nov2017_V32_MC_Uncertainty_AK4PFPuppi.junc.txt",
        "* * tests/samples/Autumn18_V8_MC_UncertaintySources_AK4PFchs.junc.txt",
        "* * tests/samples/Spring16_25nsV10_MC_SF_AK4PFPuppi.jersf.txt",
        "* * tests/samples/Autumn18_V7b_MC_SF_AK8PFchs.jersf.txt.gz",
        "* * tests/samples/Fall17_17Nov2017_V32_MC_L2Relative_AK4Calo.jec.txt.gz",
        "* * tests/samples/Fall17_17Nov2017_V32_MC_L1JPTOffset_AK4JPT.jec.txt.gz",
        "* * tests/samples/Fall17_17Nov2017B_V32_DATA_L2Relative_AK4Calo.txt.gz",
        "* * tests/samples/Autumn18_V7b_DATA_SF_AK4PF.jersf.txt",
        "* * tests/samples/Autumn18_RunC_V19_DATA_L2Relative_AK8PFchs.jec.txt.gz",
        "* * tests/samples/Autumn18_RunA_V19_DATA_L2Relative_AK4Calo.jec.txt"
    ])
    extractor.finalize()

    evaluator = extractor.make_evaluator()

    counts, test_eta, test_pt = dummy_jagged_eta_pt()
    
    jec_out = evaluator['testJECFall17_17Nov2017_V32_MC_L2Relative_AK4PFPuppi'](test_eta,test_pt)

    print(evaluator['testJECFall17_17Nov2017_V32_MC_L2Relative_AK4PFPuppi'])
    
    jec_out = evaluator['Summer16_07Aug2017_V11_L1fix_MC_L2Relative_AK4PFchs'](test_eta,test_pt)
    
    print(evaluator['Summer16_07Aug2017_V11_L1fix_MC_L2Relative_AK4PFchs'])

    jersf = evaluator['Spring16_25nsV10_MC_SF_AK4PFPuppi']
    
    print(evaluator['Spring16_25nsV10_MC_SF_AK4PFPuppi'])
    
    junc_out = evaluator['Fall17_17Nov2017_V32_MC_Uncertainty_AK4PFPuppi'](test_eta,test_pt)

    print(evaluator['Fall17_17Nov2017_V32_MC_Uncertainty_AK4PFPuppi'])
    
    assert('Autumn18_V8_MC_UncertaintySources_AK4PFchs_AbsoluteScale' in evaluator.keys())
    junc_out = evaluator['Autumn18_V8_MC_UncertaintySources_AK4PFchs_AbsoluteScale'](test_eta,test_pt)
    print(evaluator['Autumn18_V8_MC_UncertaintySources_AK4PFchs_AbsoluteScale'])

def test_jec_txt_effareas():
    extractor = lookup_tools.extractor()
    extractor.add_weight_sets(['* * tests/samples/photon_id.ea.txt'])
    extractor.finalize()

    evaluator = extractor.make_evaluator()
    
    counts, test_eta, test_pt = dummy_jagged_eta_pt()

    ch_out = evaluator['photon_id_EA_CHad'](test_eta)
    print(evaluator['photon_id_EA_CHad'])

    nh_out = evaluator['photon_id_EA_NHad'](test_eta)
    print(evaluator['photon_id_EA_NHad'])

    ph_out = evaluator['photon_id_EA_Pho'](test_eta)
    print(evaluator['photon_id_EA_Pho'])

def test_rochester():
    rochester_data = lookup_tools.txt_converters.convert_rochester_file('tests/samples/RoccoR2018.txt.gz',loaduncs=True)
    rochester = lookup_tools.rochester_lookup.rochester_lookup(rochester_data)

    # to test 1-to-1 agreement with official Rochester requires loading C++ files
    # instead, preload the correct scales in the sample directory
    # the script tests/samples/rochester/build_rochester.py produces these
    official_data_k = np.load('tests/samples/nano_dimuon_rochester.npy')
    official_data_err = np.load('tests/samples/nano_dimuon_rochester_err.npy')
    official_mc_k = np.load('tests/samples/nano_dy_rochester.npy')
    official_mc_err = np.load('tests/samples/nano_dy_rochester_err.npy')
    mc_rand = np.load('tests/samples/nano_dy_rochester_rand.npy')

    # test against nanoaod
    events = NanoEvents.from_file(os.path.abspath('tests/samples/nano_dimuon.root'))

    data_k = rochester.kScaleDT(events.Muon.charge, events.Muon.pt, events.Muon.eta, events.Muon.phi)
    assert(all(np.isclose(data_k.flatten(), official_data_k)))
    data_err = rochester.kScaleDTerror(events.Muon.charge, events.Muon.pt, events.Muon.eta, events.Muon.phi)
    data_err = np.array(data_err.flatten(), dtype=float)
    assert(all(np.isclose(data_err, official_data_err, atol=1e-8)))

    # test against mc
    events = NanoEvents.from_file(os.path.abspath('tests/samples/nano_dy.root'))

    hasgen = ~np.isnan(events.Muon.matched_gen.pt.fillna(np.nan))
    mc_rand = JaggedArray.fromoffsets(hasgen.offsets, mc_rand)
    mc_kspread = rochester.kSpreadMC(events.Muon.charge[hasgen], events.Muon.pt[hasgen], events.Muon.eta[hasgen], events.Muon.phi[hasgen],
                                     events.Muon.matched_gen.pt[hasgen])
    mc_ksmear = rochester.kSmearMC(events.Muon.charge[~hasgen], events.Muon.pt[~hasgen], events.Muon.eta[~hasgen], events.Muon.phi[~hasgen],
                                   events.Muon.nTrackerLayers[~hasgen], mc_rand[~hasgen])
    mc_k = np.ones_like(events.Muon.pt.flatten())
    mc_k[hasgen.flatten()] = mc_kspread.flatten()
    mc_k[~hasgen.flatten()] = mc_ksmear.flatten()
    assert(all(np.isclose(mc_k, official_mc_k)))

    mc_errspread = rochester.kSpreadMCerror(events.Muon.charge[hasgen], events.Muon.pt[hasgen], events.Muon.eta[hasgen], events.Muon.phi[hasgen],
                                            events.Muon.matched_gen.pt[hasgen])
    mc_errsmear = rochester.kSmearMCerror(events.Muon.charge[~hasgen], events.Muon.pt[~hasgen], events.Muon.eta[~hasgen], events.Muon.phi[~hasgen],
                                          events.Muon.nTrackerLayers[~hasgen], mc_rand[~hasgen])
    mc_err = np.ones_like(events.Muon.pt.flatten())
    mc_err[hasgen.flatten()] = mc_errspread.flatten()
    mc_err[~hasgen.flatten()] = mc_errsmear.flatten()
    assert(all(np.isclose(mc_err, official_mc_err, atol=1e-8)))




from .lookup_base import lookup_base

from ..util import awkward
from ..util import numpy
from ..util import numpy as np
from copy import deepcopy

from scipy.special import erf
from numpy import sqrt, log, log10, exp, abs
from numpy import maximum as max
from numpy import minimum as min
from numpy import power as pow


def wrap_formula(fstr, varlist):
    """
        Convert function string to python function
        Supports only simple math for now
    """
    val = fstr
    try:
        val = float(fstr)
        fstr = 'np.full_like(%s,%f)' % (varlist[0], val)
    except ValueError:
        val = fstr
    lstr = "lambda %s: %s" % (",".join(varlist), fstr)
    func = eval(lstr)
    return func


def masked_bin_eval(dim1_indices, dimN_bins, dimN_vals):
    dimN_indices = np.empty_like(dim1_indices)
    for i in np.unique(dim1_indices):
        idx = np.where(dim1_indices == i)
        dimN_indices[idx] = np.clip(np.searchsorted(dimN_bins[i],
                                                    dimN_vals[idx],
                                                    side='right') - 1,
                                    0, len(dimN_bins[i]) - 2)
    return dimN_indices


# idx_in is a tuple of indices in increasing jaggedness
# idx_out is a list of flat indices
def flatten_idxs(idx_in, jaggedarray):
    """
        This provides a faster way to convert between tuples of
        jagged indices and flat indices in a jagged array's contents
    """
    if len(idx_in) == 0:
        return np.array([], dtype=np.int)
    idx_out = jaggedarray.starts[idx_in[0]]
    if len(idx_in) == 1:
        pass
    elif len(idx_in) == 2:
        idx_out += idx_in[1]
    else:
        raise Exception('jme_standard_function only works for two binning dimensions!')

    good_idx = (idx_out < jaggedarray.content.size)
    if((~good_idx).any()):
        input_idxs = tuple([idx_out[~good_idx]] +
                           [idx_in[i][~good_idx] for i in range(len(idx_in))])
        raise Exception('Calculated invalid index {} for'
                        ' array with length {}'.format(np.vstack(input_idxs),
                                                       jaggedarray.content.size))

    return idx_out


class jme_standard_function(lookup_base):
    """
        This class defines a lookup table for jet energy corrections and resolutions.
        The JEC and JER values can be looked up with a call as follows:
        jerc_lut = jme_standard_function()
        jercs = jerc_lut(JetProperty1=jet.property1,...)
        "jercs" will be of the same shape as the input jet properties.
        The list of required jet properties are given in jersf_lut.signature
    """
    def __init__(self, formula, bins_and_orders, clamps_and_vars, parms_and_orders):
        """
            The constructor takes the output of the "convert_jec(jr)_txt_file"
            text file converter, which returns a formula, bins, and parameter values.
        """
        super(jme_standard_function, self).__init__()
        self._dim_order = bins_and_orders[1]
        self._bins = bins_and_orders[0]
        self._eval_vars = clamps_and_vars[2]
        self._eval_clamp_mins = clamps_and_vars[0]
        self._eval_clamp_maxs = clamps_and_vars[1]
        self._parm_order = parms_and_orders[1]
        self._parms = parms_and_orders[0]
        self._formula_str = formula
        self._formula = wrap_formula(formula, self._parm_order + self._eval_vars)

        for binname in self._dim_order[1:]:
            binsaslists = self._bins[binname].tolist()
            self._bins[binname] = [np.array(bins) for bins in binsaslists]

        # get the jit to compile if we've got more than one bin dim
        if len(self._dim_order) > 1:
            masked_bin_eval(np.array([0, 0]), self._bins[self._dim_order[1]], np.array([0.0, 0.0]))

        # compile the formula
        argsize = len(self._parm_order) + len(self._eval_vars)
        some_ones = tuple([50 * np.ones(argsize) for i in range(argsize)])
        _ = self._formula(*some_ones)

        self._signature = deepcopy(self._dim_order)
        for eval in self._eval_vars:
            if eval not in self._signature:
                self._signature.append(eval)
        self._dim_args = {self._dim_order[i]: i for i in range(len(self._dim_order))}
        self._eval_args = {}
        for i, argname in enumerate(self._eval_vars):
            self._eval_args[argname] = i + len(self._dim_order)
            if argname in self._dim_args.keys():
                self._eval_args[argname] = self._dim_args[argname]

    def _evaluate(self, *args):
        """ jec/jer = f(args) """
        bin_vals = {argname: args[self._dim_args[argname]] for argname in self._dim_order}
        eval_vals = {argname: args[self._eval_args[argname]] for argname in self._eval_vars}

        # lookup the bins that we care about
        dim1_name = self._dim_order[0]
        dim1_indices = np.clip(np.searchsorted(self._bins[dim1_name],
                                               bin_vals[dim1_name],
                                               side='right') - 1,
                               0, self._bins[dim1_name].size - 2)
        bin_indices = [dim1_indices]
        for binname in self._dim_order[1:]:
            bin_indices.append(masked_bin_eval(bin_indices[0],
                                               self._bins[binname],
                                               bin_vals[binname]))

        bin_tuple = tuple(bin_indices)

        # get clamp values and clip the inputs
        eval_values = []
        for eval_name in self._eval_vars:
            clamp_mins = None
            if self._eval_clamp_mins[eval_name].content.size == 1:
                clamp_mins = self._eval_clamp_mins[eval_name].content[0]
            else:
                idxs = flatten_idxs(bin_tuple, self._eval_clamp_mins[eval_name])
                clamp_mins = self._eval_clamp_mins[eval_name].content[idxs]
                if isinstance(clamp_mins, awkward.JaggedArray):
                    if clamp_mins.content.size == 1:
                        clamp_mins = clamp_mins.content[0]
                    else:
                        clamp_mins = clamp_mins.flatten()
            clamp_maxs = None
            if self._eval_clamp_maxs[eval_name].content.size == 1:
                clamp_maxs = self._eval_clamp_maxs[eval_name].content[0]
            else:
                idxs = flatten_idxs(bin_tuple, self._eval_clamp_maxs[eval_name])
                clamp_maxs = self._eval_clamp_maxs[eval_name].content[idxs]
                if isinstance(clamp_maxs, awkward.JaggedArray):
                    if clamp_maxs.content.size == 1:
                        clamp_maxs = clamp_maxs.content[0]
                    else:
                        clamp_maxs = clamp_maxs.flatten()
            eval_values.append(np.clip(eval_vals[eval_name], clamp_mins, clamp_maxs))

        # get parameter values
        parm_values = []
        if len(self._parms) > 0:
            idxs = flatten_idxs(bin_tuple, self._parms[0])
            parm_values = [parm.content[idxs] for parm in self._parms]

        return self._formula(*tuple(parm_values + eval_values))

    @property
    def signature(self):
        """ the list of all needed jet properties to be passed as kwargs to this lookup """
        return self._signature

    def __repr__(self):
        out = 'binned dims: %s\n' % (self._dim_order)
        out += 'eval vars  : %s\n' % (self._eval_vars)
        out += 'parameters : %s\n' % (self._parm_order)
        out += 'formula    : %s\n' % (self._formula_str)
        out += 'signature  : (%s)\n' % (','.join(self._signature))
        return out



from __future__ import print_function, division

import sys
import pytest
import coffea
from coffea.analysis_objects import JaggedCandidateArray, JaggedTLorentzVectorArray
import uproot
import uproot_methods
from coffea.util import awkward
from coffea.util import numpy as np

from dummy_distributions import dummy_four_momenta, gen_reco_TLV

if sys.platform.startswith("win"):
    pytest.skip("skipping tests that only function in linux", allow_module_level=True)

def test_analysis_objects():
    counts, px, py, pz, energy = dummy_four_momenta()
    thep4 = np.stack((px,py,pz,energy)).T
    
    #test JaggedTLorentzVectorArray
    tlva1 = uproot_methods.TLorentzVectorArray(px,py,pz,energy)
    tlva2 = uproot_methods.TLorentzVectorArray(thep4[:,0],thep4[:,1],
                                               thep4[:,2],thep4[:,3])
    jtlva1 = JaggedTLorentzVectorArray.fromcounts(counts,tlva1)
    jtlva2 = JaggedTLorentzVectorArray.fromcounts(counts,tlva2)
    
    jtlva1_selection1 = jtlva1[jtlva1.counts > 0]
    jtlva1_selection2 = jtlva1_selection1[jtlva1_selection1.pt > 5]

    jtlva2_selection1 = jtlva2[jtlva2.counts > 0]
    jtlva2_selection2 = jtlva1_selection1[jtlva2_selection1.pt > 5]
    
    diffx = np.abs(jtlva1.x - jtlva2.x)
    diffy = np.abs(jtlva1.y - jtlva2.y)
    diffz = np.abs(jtlva1.z - jtlva2.z)
    difft = np.abs(jtlva1.t - jtlva2.t)
    assert (diffx < 1e-8).flatten().all()
    assert (diffy < 1e-8).flatten().all()
    assert (diffz < 1e-8).flatten().all()
    assert (difft < 1e-8).flatten().all()

    #test JaggedCandidateArray
    jca1 = JaggedCandidateArray.candidatesfromcounts(counts,p4=thep4)
    jca2 = JaggedCandidateArray.candidatesfromcounts(counts,p4=thep4)
    assert( (jca1.offsets == jca2.offsets).all() )

    addon1 = jca1.zeros_like()
    addon2 = jca2.ones_like()
    jca1['addon'] = addon1
    jca2['addon'] = addon2

    jca1.add_attributes(addonFlat=addon1.flatten(),addonJagged=addon1)
    
    diffm = np.abs(jca1.p4.mass - jca2.p4.mass)
    assert( (jca1.offsets == jca2.offsets).all() )
    diffpt = np.abs(jca1.p4.pt - jca2.p4.pt)
    assert( (jca1.offsets == jca2.offsets).all() )
    eta2 = jca2.p4.eta
    eta1 = jca1.p4.eta
    print (np.sum(eta1.counts),np.sum(eta2.counts))
    diffeta_temp = np.abs(eta1 - eta2)
    diffeta = np.abs(jca1.p4.eta - jca2.p4.eta)
    assert( (jca1.offsets == jca2.offsets).all() )
    assert (diffm < 1e-8).flatten().all()
    assert (diffpt < 1e-8).flatten().all()
    assert (diffeta < 1e-8).flatten().all()
    
    #test fast functions
    fastfs = ['pt','eta','phi','mass']
    for func in fastfs:
        func1 = getattr(jca1,func)        
        func2 = getattr(jca1.p4,func)
        dfunc = np.abs(func1 - func2)
        assert (dfunc < 1e-8).flatten().all()

    adistinct = jca1.distincts()
    apair = jca1.pairs()
    across = jca1.cross(jca2)
    acrossn = jca1.cross(jca2, nested=True)
    achoose2 = jca1.choose(2)
    achoose3 = jca1.choose(3)
    
    assert 'p4' in adistinct.columns
    assert 'p4' in apair.columns
    assert 'p4' in across.columns
    assert 'p4' in acrossn.columns
    assert 'p4' in achoose2.columns
    assert 'p4' in achoose3.columns
    
    admsum = (adistinct.i0.p4 + adistinct.i1.p4).mass
    apmsum = (apair.i0.p4 + apair.i1.p4).mass
    acmsum = (across.i0.p4 + across.i1.p4).mass
    ach3msum = (achoose3.i0.p4 + achoose3.i1.p4 + achoose3.i2.p4).mass
    diffadm = np.abs(adistinct.p4.mass - admsum)
    diffapm = np.abs(apair.p4.mass - apmsum)
    diffacm = np.abs(across.p4.mass - acmsum)
    diffachm = np.abs(achoose2.p4.mass - admsum)
    diffach3m = np.abs(achoose3.p4.mass - ach3msum)
    
    assert (diffadm < 1e-8).flatten().all()
    assert (diffapm < 1e-8).flatten().all()
    assert (diffacm < 1e-8).flatten().all()
    assert (diffachm < 1e-8).flatten().all()
    assert (diffach3m < 1e-8).flatten().all()

    selection11 = jca1[jca1.counts > 0]
    selection12 = selection11[selection11.p4.pt > 5]
    
    selection21 = jca2[jca2.counts > 0]
    selection22 = selection21[selection21.p4.pt > 5]

    diffcnts = selection12.counts - jtlva1_selection2.counts
    diffm = np.abs(selection12.p4.mass - jtlva1_selection2.mass)
    diffaddon = selection12.addon - selection22.addon
    assert (diffcnts == 0).flatten().all()
    assert (diffm < 1e-8).flatten().all()
    assert (diffaddon == -1).flatten().all()

    #test gen-reco matching
    gen, reco = gen_reco_TLV()
    flat_gen = gen.flatten()
    gen_px,gen_py,gen_pz,gen_e = flat_gen.x,flat_gen.y,flat_gen.z,flat_gen.t
    flat_reco = reco.flatten()
    reco_px,reco_py,reco_pz,reco_e = flat_reco.x,flat_reco.y,flat_reco.z,flat_reco.t
    jca_gen = JaggedCandidateArray.candidatesfromcounts(gen.counts,
                                                        px=gen_px,py=gen_py,
                                                        pz=gen_pz,energy=gen_e)
    jca_reco = JaggedCandidateArray.candidatesfromcounts(reco.counts,
                                                         px=reco_px,py=reco_py,
                                                         pz=reco_pz,energy=reco_e)
    print('gen eta: ', jca_gen.p4.eta,'\n gen phi:', jca_gen.p4.phi)
    print('reco eta: ', jca_reco.p4.eta,'\n reco phi:', jca_reco.p4.phi)
    match_mask = jca_reco.match(jca_gen, deltaRCut=0.3)
    print('match mask: ',  match_mask)
    fast_match_mask = jca_reco.fastmatch(jca_gen, deltaRCut=0.3)
    print('fastmatch mask: ', fast_match_mask)
    assert((match_mask == fast_match_mask).all().all())
    print('arg matches: ',jca_reco.argmatch(jca_gen,deltaRCut=0.3))
    argmatch_nocut = jca_gen.argmatch(jca_reco).flatten()
    argmatch_dr03  = jca_gen.argmatch(jca_reco, deltaRCut=0.3).flatten()
    argmatch_dr03_dpt01 = jca_gen.argmatch(jca_reco, deltaRCut=0.3, deltaPtCut=0.1).flatten()
    assert (argmatch_nocut.size==5)
    assert (argmatch_dr03[argmatch_dr03 != -1].size==3)
    assert (argmatch_dr03_dpt01[argmatch_dr03_dpt01 != -1].size==2)
    assert (jca_gen.match(jca_reco,deltaRCut=0.3).flatten().flatten().sum()==3)
    assert (jca_gen.match(jca_reco,deltaRCut=0.3,deltaPtCut=0.1).flatten().flatten().sum()==2)

    # test various four-momentum constructors
    ptetaphiE_test = JaggedCandidateArray.candidatesfromcounts(jca_reco.counts,
                                                               pt=jca_reco.pt,
                                                               eta=jca_reco.eta,
                                                               phi=jca_reco.phi,
                                                               energy=jca_reco.p4.energy)

    pxpypzM_test = JaggedCandidateArray.candidatesfromcounts(jca_reco.counts,
                                                             px=jca_reco.p4.x,
                                                             py=jca_reco.p4.y,
                                                             pz=jca_reco.p4.z,
                                                             mass=jca_reco.mass)

    ptphipzE_test = JaggedCandidateArray.candidatesfromcounts(jca_reco.counts,
                                                              pt=jca_reco.pt,
                                                              phi=jca_reco.phi,
                                                              pz=jca_reco.p4.z,
                                                              energy=jca_reco.p4.energy)

    pthetaphiE_test = JaggedCandidateArray.candidatesfromcounts(jca_reco.counts,
                                                                p=jca_reco.p4.p,
                                                                theta=jca_reco.p4.theta,
                                                                phi=jca_reco.phi,
                                                                energy=jca_reco.p4.energy)




import numpy
import uproot
import awkward
from .methods import collection_methods
from .util import _mixin
from ..util import _hex, _ascii


class NanoCollection(awkward.VirtualArray):
    r'''A NanoAOD collection

    NanoAOD collections are collections of branches formed by name, where:

    - one branch exists named ``name`` and no branches start with ``name_``, interpreted as a single flat array;
    - one branch exists named ``name``, one named ``n{name}``, and no branches start with ``name_``, interpreted as a single jagged array;
    - no branch exists named ``{name}`` and many branches start with ``name_*``, interpreted as a flat table; or
    - one branch exists named ``n{name}`` and many branches start with ``name_*``, interpreted as a jagged table.

    '''
    _mixin_registry = {}

    @classmethod
    def _get_mixin(cls, methods, awkwardtype=None):
        if awkwardtype is None:
            awkwardtype = NanoCollection
        if issubclass(awkwardtype, NanoCollection) and awkwardtype is not NanoCollection:
            raise ValueError("Can only mixin with NanoCollection base or awkward type, not derived")
        key = (methods, awkwardtype)
        if key not in NanoCollection._mixin_registry:
            if methods is None:
                NanoCollection._mixin_registry[key] = awkwardtype
            else:
                NanoCollection._mixin_registry[key] = _mixin(methods, awkwardtype)
        return NanoCollection._mixin_registry[key]

    @classmethod
    def _get_methods(cls):
        for k, v in NanoCollection._mixin_registry.items():
            if v is cls:
                return k[0]
        raise RuntimeError("Unregistered mixin detected")

    @classmethod
    def from_arrays(cls, arrays, name, methods=None):
        '''Build from dictionary of VirtualArrays

        Parameters
        ----------
            arrays : dict
                A mapping from branch name to flat VirtualArray
            name : str
                The name of the collection (see class documentation for interpretation)
            methods : class, optional
                A class deriving from `awkward.array.objects.Methods` that implements additional mixins

        Returns a NanoCollection object, possibly mixed in with methods.
        '''
        outtype = cls._get_mixin(methods)
        jagged = 'n' + name in arrays.keys()
        columns = {k[len(name) + 1:]: arrays[k] for k in arrays.keys() if k.startswith(name + '_')}
        if len(columns) == 0:
            # single-item collection, just forward lazy array
            if name not in arrays.keys():
                raise RuntimeError('Could not find collection %s in dataframe' % name)
            array = outtype.__new__(outtype)
            array.__dict__ = arrays[name].__dict__
            array.__doc__ = arrays[name].__doc__
            if jagged:
                counts = arrays['n' + name]
                out = outtype(
                    outtype._lazyjagged,
                    (name, counts, array, methods),
                    type=awkward.type.ArrayType(len(counts), float('inf'), array.type.to),
                )
                out.__doc__ = counts.__doc__
                array = out
            return array
        elif jagged:
            tabletype = awkward.type.TableType()
            for k, array in columns.items():
                tabletype[k] = array.type.to
            counts = arrays['n' + name]
            out = outtype(
                outtype._lazyjagged,
                (name, counts, columns, methods),
                type=awkward.type.ArrayType(len(counts), float('inf'), tabletype),
            )
            out.__doc__ = counts.__doc__
            return out
        else:
            Table = cls._get_mixin(methods, awkward.Table)
            table = Table.named(name)
            for k, v in columns.items():
                table.contents[k] = v
            return table

    @classmethod
    def _lazyjagged(cls, name, counts, columns, methods=None):
        offsets = awkward.JaggedArray.counts2offsets(counts.array)
        JaggedArray = cls._get_mixin(methods, awkward.JaggedArray)
        Table = cls._get_mixin(methods, awkward.Table)
        if isinstance(columns, dict):
            content = Table.named(name)
            content.type.takes = offsets[-1]
            for k, v in columns.items():
                if not isinstance(v, awkward.VirtualArray):
                    raise RuntimeError
                v.type.takes = offsets[-1]
                content.contents[k] = v
        else:
            if not isinstance(columns, awkward.VirtualArray):
                raise RuntimeError
            columns.type.takes = offsets[-1]
            content = columns
        out = JaggedArray.fromoffsets(offsets, content)
        out.__doc__ = counts.__doc__
        return out

    def _lazy_crossref(self, index, destination):
        if not isinstance(destination, NanoCollection):
            raise ValueError("Destination must be a NanoCollection")
        if not isinstance(destination.array, awkward.JaggedArray):
            raise ValueError("Cross-references imply jagged destination")
        if not isinstance(self.array, awkward.JaggedArray):
            raise NotImplementedError
        IndexedMaskedArray = destination._get_mixin(destination._get_methods(), awkward.IndexedMaskedArray)
        # repair awkward type now that we've materialized
        index.type.takes = self.array.offsets[-1]
        index = awkward.JaggedArray.fromoffsets(self.array.offsets, content=index)
        globalindex = (index + destination.array.starts).flatten()
        invalid = (index < 0).flatten()
        globalindex[invalid] = -1
        # note: parent virtual must derive from this type
        out = IndexedMaskedArray(
            globalindex,
            destination.array.content,
        )
        # useful for algorithms
        self.array.content['_xref_%s_index' % destination.rowname] = globalindex
        return out

    def _lazy_nested_crossref(self, indices, destination):
        if not isinstance(destination, NanoCollection):
            raise ValueError("Destination must be a NanoCollection")
        if not isinstance(destination.array, awkward.JaggedArray):
            raise ValueError("Cross-references imply jagged destination")
        if not isinstance(self.array, awkward.JaggedArray):
            raise NotImplementedError
        JaggedArray = destination._get_mixin(destination._get_methods(), awkward.JaggedArray)
        # repair type now that we've materialized
        for idx in indices:
            idx.type.takes = self.array.offsets[-1]
        content = numpy.zeros(len(self.array.content) * len(indices), dtype=awkward.JaggedArray.INDEXTYPE)
        for i, index in enumerate(indices):
            content[i::len(indices)] = numpy.array(index)
        globalindices = awkward.JaggedArray.fromoffsets(
            self.array.offsets,
            awkward.JaggedArray.fromoffsets(
                numpy.arange((len(self.array.content) + 1) * len(indices), step=len(indices)),
                content,
            )
        )
        globalindices = globalindices[globalindices >= 0] + destination.array.starts
        # note: parent virtual must derive from this type
        out = JaggedArray.fromoffsets(
            globalindices.content.offsets,
            content=destination.array.content[globalindices.flatten().flatten()]
        )
        return out

    def _getcolumn(self, key):
        name, _, columns, _ = self._args
        if key not in columns:
            # This function is only meant for use in methods' _finalize() while
            # all columns are still virtual. Missing arrays are a sign of an incompatible
            # file or missing preloaded columns. This triggers only if the missing column is accessed.
            def nonexistentarray():
                raise RuntimeError("There was an attempt to read the nonexistent array: %s_%s" % (name, key))
            return awkward.VirtualArray(nonexistentarray)
        return columns[key]

    def __setitem__(self, key, value):
        if self.ismaterialized:
            super(NanoCollection, self).__setitem__(key, value)
        _, _, columns, _ = self._args
        columns[key] = value
        self._type.to.to[key] = value.type.to

    def __delitem__(self, key):
        if self.ismaterialized:
            super(NanoCollection, self).__delitem__(key)
        _, _, columns, _ = self._args
        if key in columns:
            del columns[key]
            del self._type.to.to[key]


class NanoEvents(awkward.Table):
    '''Top-level awkward array representing NanoAOD events chunk

    The default interpretation of the various collections is governed by `coffea.nanoaod.methods.collection_methods`,
    but can optionally be overriden by passing a custom mapping in the various ``from_*`` constructors.
    '''
    @classmethod
    def from_arrays(cls, arrays, methods=None, metadata=None):
        '''Build NanoEvents from a dictionary of arrays

        Parameters
        ----------
            arrays : dict
                A mapping from branch name to flat numpy array or awkward VirtualArray
            methods : dict, optional
                A mapping from collection name to class deriving from `awkward.array.objects.Methods`
                that implements additional mixins
            metadata : dict, optional
                Arbitrary metadata to embed in this NanoEvents table

        Returns a NanoEvents object
        '''
        arrays = dict(arrays)
        for k in arrays:
            if isinstance(arrays[k], awkward.VirtualArray):
                pass
            elif isinstance(arrays[k], numpy.ndarray):
                value = arrays[k]
                arrays[k] = awkward.VirtualArray(lambda: value, type=awkward.type.ArrayType(len(arrays[k]), arrays[k].dtype))
                print(arrays[k])
            else:
                raise ValueError("The array %s : %r is not a valid type" % (k, arrays[k]))
        events = cls.named('event')
        collections = {k.split('_')[0] for k in arrays.keys()}
        collections -= {k for k in collections if k.startswith('n') and k[1:] in collections}
        allmethods = {}
        allmethods.update(collection_methods)
        if methods is not None:
            allmethods.update(methods)
        for name in collections:
            methods = allmethods.get(name, None)
            events.contents[name] = NanoCollection.from_arrays(arrays, name, methods)

        for name in events.columns:
            # soft hasattr via type, to prevent materialization
            if hasattr(type(events[name]), '_finalize'):
                events.contents[name]._finalize(name, events)

        events.metadata = metadata if metadata is not None else {}
        return events

    @classmethod
    def from_file(cls, file, treename=b'Events', entrystart=None, entrystop=None, cache=None, methods=None, metadata=None):
        '''Build NanoEvents directly from ROOT file

        Parameters
        ----------
            file : str or uproot.rootio.ROOTDirectory
                The filename or already opened file using e.g. ``uproot.open()``
            treename : str, optional
                Name of the tree to read in the file, defaults to ``Events``
            entrystart : int, optional
                Start at this entry offset in the tree (default 0)
            entrystop : int, optional
                Stop at this entry offset in the tree (default end of tree)
            cache : dict, optional
                A dict-like interface to a cache object, in which any materialized virtual arrays will be kept
            methods : dict, optional
                A mapping from collection name to class deriving from `awkward.array.objects.Methods`
                that implements custom additional mixins beyond the defaults provided.
            metadata : dict, optional
                Arbitrary metadata to embed in this NanoEvents table

        Returns a NanoEvents object
        '''
        if cache is None:
            cache = {}
        if not isinstance(file, uproot.rootio.ROOTDirectory):
            file = uproot.open(file)
        tree = file[treename]
        entrystart, entrystop = uproot.tree._normalize_entrystartstop(tree.numentries, entrystart, entrystop)
        arrays = {}
        for bname in tree.keys():
            interpretation = uproot.interpret(tree[bname])
            if isinstance(interpretation, uproot.asjagged):
                virtualtype = awkward.type.ArrayType(float('inf'), interpretation.content.type)
            else:
                virtualtype = awkward.type.ArrayType(entrystop - entrystart, interpretation.type)
            array = awkward.VirtualArray(
                tree[bname].array,
                (),
                {'entrystart': entrystart, 'entrystop': entrystop, 'flatten': True},
                type=virtualtype,
                persistentkey=';'.join(str(x) for x in (_hex(file._context.uuid), _ascii(treename), entrystart, entrystop, _ascii(bname))),
                cache=cache,
            )
            array.__doc__ = tree[bname].title.decode('ascii')
            arrays[bname.decode('ascii')] = array
        out = cls.from_arrays(arrays, methods=methods, metadata=metadata)
        out._cache = cache
        return out

    def tolist(self):
        '''Overriden to raise an exception.  Do not call'''
        raise NotImplementedError("NanoEvents cannot be rendered as a list due to cyclic cross-references")

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        '''Overriden to raise an exception.  Do not call'''
        raise NotImplementedError("NanoEvents cannot be broadcast due to cyclic cross-references")

    def __getitem__(self, where):
        '''Overriden to keep metadata sticky'''
        out = super(NanoEvents, self).__getitem__(where)
        if isinstance(out, NanoEvents):
            out.metadata = self.metadata
            if hasattr(self, '_cache'):
                out._cache = self._cache
        return out

    @property
    def materialized(self):
        '''Set of columns or branches that were materialized

        This is only available when constructed from the file constructor
        '''
        if not hasattr(self, '_cache'):
            raise RuntimeError("NanoEvents.materialized only available if constructed with from_file")
        return set(k.split(';')[-1] for k in self._cache.keys())



from .common import LorentzVector
import awkward


class FatJet(LorentzVector):
    '''NanoAOD large radius jet object'''
    subjetmap = {'FatJet': 'SubJet'}  # V6 has 'GenJetAK8': 'SubGenJetAK8', maybe better to put in generator.py
    LOOSE = 0
    'jetId bit position'
    TIGHT = 1
    'jetId bit position'
    TIGHTLEPVETO = 2
    'jetId bit position'

    def _finalize(self, name, events):
        subjets = self.subjetmap.get(name, None)
        if subjets is not None and subjets in events.columns:
            subjets = events[subjets]
            embedded_subjets = type(subjets)(
                self._lazy_nested_crossref,
                args=([self._getcolumn('subJetIdx1'), self._getcolumn('subJetIdx2')], subjets),
                type=awkward.type.ArrayType(float('inf'), float('inf'), subjets.type.to.to),
            )
            embedded_subjets.__doc__ = subjets.__doc__
            self['subjets'] = embedded_subjets
            del self['subJetIdx1']
            del self['subJetIdx2']

    @property
    def isLoose(self):
        '''Returns a boolean array marking loose jets according to jetId index'''
        return (self.jetId & (1 << self.LOOSE)).astype(bool)

    @property
    def isTight(self):
        '''Returns a boolean array marking tight jets according to jetId index'''
        return (self.jetId & (1 << self.TIGHT)).astype(bool)

    @property
    def isTightLeptonVeto(self):
        '''Returns a boolean array marking tight jets with explicit lepton veto according to jetId index'''
        return (self.cutBasedBitmap & (1 << self.TIGHTLEPVETO)).astype(bool)


class Jet(LorentzVector):
    '''NanoAOD narrow radius jet object'''
    _enable_genjet = False
    'Set to true if using NanoAODv6 or newer (v5 had a bug in the mapping)'
    LOOSE = 0
    'jetId bit position'
    TIGHT = 1
    'jetId bit position'
    TIGHTLEPVETO = 2
    'jetId bit position'

    def _finalize(self, name, events):
        if 'Electron' in events.columns:
            electrons = events['Electron']
            reftype = awkward.type.ArrayType(float('inf'), float('inf'), electrons.type.to.to)
            reftype.check = False
            embedded_electrons = type(electrons)(
                self._lazy_nested_crossref,
                args=([self._getcolumn('electronIdx1'), self._getcolumn('electronIdx2')], electrons),
                type=reftype,
            )
            embedded_electrons.__doc__ = electrons.__doc__
            self['matched_electrons'] = embedded_electrons
            del self['electronIdx1']
            del self['electronIdx2']

        if 'Muon' in events.columns:
            muons = events['Muon']
            reftype = awkward.type.ArrayType(float('inf'), float('inf'), muons.type.to.to)
            reftype.check = False
            embedded_muons = type(muons)(
                self._lazy_nested_crossref,
                args=([self._getcolumn('muonIdx1'), self._getcolumn('muonIdx2')], muons),
                type=reftype,
            )
            embedded_muons.__doc__ = muons.__doc__
            self['matched_muons'] = embedded_muons
            del self['muonIdx1']
            del self['muonIdx2']

        if self._enable_genjet and 'GenJet' in events.columns:
            genjet = events['GenJet']
            reftype = awkward.type.ArrayType(float('inf'), awkward.type.OptionType(genjet.type.to.to))
            reftype.check = False
            embedded_genjet = type(genjet)(
                self._lazy_crossref,
                args=(self._getcolumn('genJetIdx'), genjet),
                type=reftype,
            )
            embedded_genjet.__doc__ = genjet.__doc__
            self['matched_gen'] = embedded_genjet
            del self['genJetIdx']

        self.type.check = False

    @property
    def isLoose(self):
        '''Returns a boolean array marking loose jets according to jetId index'''
        return (self.jetId & (1 << self.LOOSE)).astype(bool)

    @property
    def isTight(self):
        '''Returns a boolean array marking tight jets according to jetId index'''
        return (self.jetId & (1 << self.TIGHT)).astype(bool)

    @property
    def isTightLeptonVeto(self):
        '''Returns a boolean array marking tight jets with explicit lepton veto according to jetId index'''
        return (self.cutBasedBitmap & (1 << self.TIGHTLEPVETO)).astype(bool)



import ROOT
ROOT.gROOT.SetBatch(True)
ROOT.PyConfig.IgnoreCommandLineOptions = True
ROOT.gInterpreter.LoadFile('BTagCalibrationStandalone.cpp')

from functools import partial
import numpy
from coffea.btag_tools import BTagScaleFactor


def stdvec(l):
    out = ROOT.std.vector('string')()
    for item in l:
        out.push_back(item)
    return out


def makesf(btagReader):
    def btv_sf(syst, flavor, abseta, pt, discr=None):
        def wrap(flavor, abseta, pt, discr):
            if flavor == 5:
                btvflavor = ROOT.BTagEntry.FLAV_B
            elif flavor == 4:
                btvflavor = ROOT.BTagEntry.FLAV_C
            elif flavor == 0:
                btvflavor = ROOT.BTagEntry.FLAV_UDSG
            else:
                raise ValueError
            if discr is None:
                return btagReader.eval_auto_bounds(syst, btvflavor, abseta, pt)
            return btagReader.eval_auto_bounds(syst, btvflavor, abseta, pt, discr)
        return numpy.vectorize(wrap, otypes='d')(flavor, abseta, pt, discr)
    return btv_sf


def validate_btag(filename, btagtype, etamax):
    btagData = ROOT.BTagCalibration(btagtype, filename)
    npts = 10000
    flavor = numpy.full(npts, 5)
    abseta = numpy.random.uniform(0, etamax, size=npts)
    pt = numpy.random.exponential(50, size=npts) + numpy.random.exponential(20, size=npts)
    pt = numpy.maximum(20.1, pt)
    discr = numpy.random.rand(npts)

    coffea_sf = BTagScaleFactor(filename, BTagScaleFactor.RESHAPE, 'iterativefit', keep_df=True)
    btagReader = ROOT.BTagCalibrationReader(ROOT.BTagEntry.OP_RESHAPING, 'central', stdvec(['up_jes', 'down_jes']))
    btagReader.load(btagData, ROOT.BTagEntry.FLAV_B, 'iterativefit')
    btv_sf = makesf(btagReader)

    for syst in ['central', 'up_jes', 'down_jes']:
        csf = coffea_sf.eval(syst, flavor, abseta, pt, discr)
        bsf = btv_sf(syst, flavor, abseta, pt, discr)
        print(abs(csf - bsf).max())

    flavor = numpy.random.choice([0, 4, 5], size=npts)
    coffea_sf = BTagScaleFactor(filename, BTagScaleFactor.TIGHT, 'comb,mujets,incl', keep_df=True)
    btagReader = ROOT.BTagCalibrationReader(ROOT.BTagEntry.OP_TIGHT, 'central', stdvec(['up', 'down']))
    btagReader.load(btagData, ROOT.BTagEntry.FLAV_B, 'comb')
    btagReader.load(btagData, ROOT.BTagEntry.FLAV_C, 'mujets')
    btagReader.load(btagData, ROOT.BTagEntry.FLAV_UDSG, 'incl')
    btv_sf = makesf(btagReader)

    for syst in ['central', 'up', 'down']:
        csf = coffea_sf.eval(syst, flavor, abseta, pt, discr)
        bsf = btv_sf(syst, flavor, abseta, pt, discr)
        print(abs(csf - bsf).max())


validate_btag(
    filename='DeepCSV_102XSF_V1.btag.csv',
    btagtype='DeepCSV',
    etamax=2.5,
)
validate_btag(
    filename='testBTagSF.btag.csv',
    btagtype='CSVv2',
    etamax=2.4,
)



from six import with_metaclass
from abc import ABCMeta, abstractmethod


class ProcessorABC(with_metaclass(ABCMeta)):
    '''ABC for a generalized processor

    The various data delivery mechanisms (spark, striped, uproot, uproot+futures, condor, ...)
    receive such an object and the appropriate metadata to deliver dataframes to it.
    It is expected that the entire processor object can be serializable (check with `coffea.util.save`)
    No attempt should be made to track state inside an instance of ProcessorABC, it is to be
    treated simply as a bundle of methods.  The only exception is the read-only accumulator
    property.

    Examples
    --------

    A skeleton processor::

        from coffea import hist, processor

        class MyProcessor(processor.ProcessorABC):
            def __init__(self, flag=False):
                self._flag = flag
                self._accumulator = processor.dict_accumulator({
                    "sumw": processor.defaultdict_accumulator(float),
                })

            @property
            def accumulator(self):
                return self._accumulators

            def process(self, df):
                output = self.accumulator.identity()

                # ...

                return output

            def postprocess(self, accumulator):
                return accumulator

        p = MyProcessor()

    '''
    @property
    @abstractmethod
    def accumulator(self):
        '''Read-only accumulator object

        It is up to the derived class to define this, in e.g. the initializer.
        '''
        pass

    @abstractmethod
    def process(self, df):
        '''Processes a single DataFrame

        Returns a filled accumulator object which can be initialized via self.accumulator.identity()
        '''
        pass

    @abstractmethod
    def postprocess(self, accumulator):
        '''Final processing on aggregated accumulator

        Do any final processing on the resulting accumulator object, and return it
        '''
        pass



import numpy
import awkward
from uproot_methods.classes.TLorentzVector import PtEtaPhiMassArrayMethods
from uproot_methods.classes.TVector2 import ArrayMethods as XYArrayMethods


def _memoize(obj, name, constructor):
    memoname = '_memo_' + name
    if memoname not in obj.columns:
        out = constructor(obj)
        try:
            obj[memoname] = out
        except ValueError:
            # FIXME: add a column to view of the table
            return out
    return obj[memoname]


class METVector(XYArrayMethods):
    '''Implements the usual TVector2 methods, storing minimal materialized arrays in polar coordinates'''
    def __getitem__(self, key):
        if awkward.AwkwardArray._util_isstringslice(key) and key == 'fX':
            return _memoize(self, 'fX', lambda self: self['pt'] * numpy.cos(self['phi']))
        elif awkward.AwkwardArray._util_isstringslice(key) and key == 'fY':
            return _memoize(self, 'fY', lambda self: self['pt'] * numpy.sin(self['phi']))
        return super(METVector, self).__getitem__(key)

    # shortcut XYArrayMethods for pt and phi
    @property
    def pt(self):
        return self['pt']

    @property
    def phi(self):
        return self['phi']


class LorentzVector(PtEtaPhiMassArrayMethods):
    '''Implements the usual TLorentzVector methods, storing minimal materialized arrays in pt-eta-phi-mass coordinates'''
    _keymap = {'fPt': 'pt', 'fEta': 'eta', 'fPhi': 'phi', 'fMass': 'mass'}

    def __getitem__(self, key):
        if awkward.AwkwardArray._util_isstringslice(key) and key in self._keymap:
            if key == 'fMass' and 'mass' not in self.columns and 'fMass' not in self.columns:
                return _memoize(self, 'fMass', lambda self: self['pt'].zeros_like())
            elif 'fMass' not in self.columns:
                return self[self._keymap[key]]
        return super(LorentzVector, self).__getitem__(key)


class Candidate(LorentzVector):
    '''Implements the usual TLorentzVector methods but also propogates sum of charges'''
    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        out = super(Candidate, self).__array_ufunc__(ufunc, method, *inputs, **kwargs)
        if ufunc is numpy.add and all(isinstance(i, Candidate) for i in inputs):
            out['charge'] = getattr(ufunc, method)(*(i['charge'] for i in inputs), **kwargs)
        # TODO else: type demotion?
        return out



from coffea import (hist,processor)

import warnings

import numpy as np

import pytest


def test_spark_imports():
    pyspark = pytest.importorskip("pyspark", minversion="2.4.1")
    
    from coffea.processor.spark.spark_executor import spark_executor
    from coffea.processor.spark.detail import (_spark_initialize,
                                               _spark_make_dfs,
                                               _spark_stop)

    spark = _spark_initialize()
    _spark_stop(spark)


def test_spark_executor():
    pyspark = pytest.importorskip("pyspark", minversion="2.4.1")
    from pyarrow.compat import guid
    
    from coffea.processor.spark.detail import (_spark_initialize,
                                               _spark_make_dfs,
                                               _spark_stop)
    from coffea.processor import run_spark_job

    import os
    import os.path as osp

    import pyspark.sql
    spark_config = pyspark.sql.SparkSession.builder \
        .appName('spark-executor-test-%s' % guid()) \
        .master('local[*]') \
        .config('spark.sql.execution.arrow.enabled','true') \
        .config('spark.executor.x509proxyname','x509_u12409') \
        .config('spark.sql.execution.arrow.maxRecordsPerBatch', 200000)

    spark = _spark_initialize(config=spark_config,log_level='ERROR',spark_progress=False)

    filelist = {'ZJets': {'files': ['file:'+osp.join(os.getcwd(),'tests/samples/nano_dy.root')], 'treename': 'Events' },
                'Data'  : {'files': ['file:'+osp.join(os.getcwd(),'tests/samples/nano_dimuon.root')], 'treename': 'Events'}
                }

    from coffea.processor.test_items import NanoTestProcessor
    from coffea.processor.spark.spark_executor import spark_executor

    columns = ['nMuon','Muon_pt','Muon_eta','Muon_phi','Muon_mass']
    proc = NanoTestProcessor(columns=columns)

    hists = run_spark_job(filelist, processor_instance=proc, executor=spark_executor, spark=spark, thread_workers=1,
                          executor_args={'file_type': 'root'})

    _spark_stop(spark)

    assert( sum(spark_executor.counts.values()) == 80 )
    assert( hists['cutflow']['ZJets_pt'] == 18 )
    assert( hists['cutflow']['ZJets_mass'] == 6 )
    assert( hists['cutflow']['Data_pt'] == 84 )
    assert( hists['cutflow']['Data_mass'] == 66 )


def test_spark_hist_adders():
    pyspark = pytest.importorskip("pyspark", minversion="2.4.1")
    
    import pandas as pd
    import pickle as pkl
    import lz4.frame as lz4f

    from coffea.util import numpy as np
    from coffea.processor.spark.spark_executor import agg_histos_raw, reduce_histos_raw
    from coffea.processor.test_items import NanoTestProcessor

    proc = NanoTestProcessor()

    one = proc.accumulator.identity()
    two = proc.accumulator.identity()
    hlist1 = [lz4f.compress(pkl.dumps(one))]
    hlist2 = [lz4f.compress(pkl.dumps(one)),lz4f.compress(pkl.dumps(two))]
    harray1 = np.array(hlist1, dtype='O')
    harray2 = np.array(hlist2, dtype='O')
    
    series1 = pd.Series(harray1)
    series2 = pd.Series(harray2)
    df = pd.DataFrame({'histos': harray2})

    # correctness of these functions is checked in test_spark_executor
    agg1 = agg_histos_raw(series1, proc, 1)
    agg2 = agg_histos_raw(series2, proc, 1)
    red = reduce_histos_raw(df, proc, 1)



from .common import METVector, LorentzVector, Candidate
from .leptons import Electron, Muon, Photon, Tau
from .jets import Jet, FatJet
from .generator import GenParticle, GenVisTau


collection_methods = {
    'CaloMET': METVector,
    'ChsMET': METVector,
    'GenMET': METVector,
    'MET': METVector,
    'METFixEE2017': METVector,
    'PuppiMET': METVector,
    'RawMET': METVector,
    'TkMET': METVector,
    # pseudo-lorentz: pt, eta, phi, mass=0
    'IsoTrack': LorentzVector,
    'SoftActivityJet': LorentzVector,
    'TrigObj': LorentzVector,
    # True lorentz: pt, eta, phi, mass
    'FatJet': FatJet,
    'GenDressedLepton': LorentzVector,
    'GenJet': LorentzVector,
    'GenJetAK8': FatJet,
    'Jet': Jet,
    'LHEPart': LorentzVector,
    'SV': LorentzVector,
    'SubGenJetAK8': LorentzVector,
    'SubJet': LorentzVector,
    # Candidate: LorentzVector + charge
    'Electron': Electron,
    'Muon': Muon,
    'Photon': Photon,
    'Tau': Tau,
    'GenVisTau': GenVisTau,
    # special
    'GenPart': GenParticle,
}

__all__ = [
    'METVector',
    'LorentzVector',
    'Candidate',
    'Electron',
    'Muon',
    'Photon',
    'Tau',
    'Jet',
    'FatJet',
    'GenParticle',
    'GenVisTau',
    'collection_methods',
]
